import numpy
import xmlrpclib
from ibidas.utils import util,cutils

_delay_import_(globals(),"ibidas","*")
_delay_import_(globals(),"..itypes","rtypes","dimpaths","dimensions")


@util.memoized
def cyconnect(url=None):
    import xmlrpclib
    if url is None:
        url = "http://localhost:9000"
    return xmlrpclib.ServerProxy("http://localhost:9000").Cytoscape


def nan_to_zero(val):
    if numpy.isnan(val):
        return 0.
    else:
        return val


class CyNetwork(object):
    def __init__(self, name=None, url=None, exists=False):
        self._server = cyconnect(url)

        networkid = None
        if exists:
            if name is None:
                assert self._server.hasCurrentNetwork(), "No current network found"
                self._networkid = self._server.getCurrentNetworkIndex()
            else:
                nlist = self._server.getNetworkList()
                nlist = [elem for elem in nlist if elem['networktitle'] == name]
                assert len(nlist) > 1, "Multiple networks with same name. Maybe select on current network (no name)?"
                self._networkid = nlist[0]['networkID']
        else:                
            if name is None:
                name = "Network"
            self._networkid = self._server.createNetwork(name)


    def _getServ(self):
        return self._server
    Direct = property(fget=_getServ)

    def _getId(self):
        return self._networkid
    Id = property(fget=_getId)

    def AddNodes(self, rep, primary=0):
        assert isinstance(rep, representor.Representor), "Rep should be a representor object"
        rep = rep.Copy()

        primary = rep.Get(primary)
        attributes = rep.Without(primary)
        attributes = attributes.Level(1)
        
        #package data if needed
        primary = primary.Level(1).Cast("bytes")
       
        #some checks
        assert len(primary.Slices) == 1, "Node identifiers should consist of only one slice"
        dims = set([primary.Slices[0].dims]) | set([aslice.dims for aslice in attributes.Slices])
        assert len(dims) == 1, "Multiple root dimensions not allowed"

        nodeids = primary.ToPython()
        self._server.createNodes(self._networkid, nodeids)
        
        for pos, slice in enumerate(attributes.Slices):
            attribute = attributes.Get(pos)
            xtype, attribute = self._prepareAttribute(attribute)

            attrmap = dict(Combine(primary, attribute).Tuple().ToPython())
            self._server.addNodeAttributes(slice.name, xtype,attrmap,False)
            
    def AddEdges(self, rep, left=0, right=1, itype=None, directed=True, addNodes=True):
        assert isinstance(rep, representor.Representor), "Rep should be a representor object"
        if itype is None:
            itype = Rep("Interaction")

        directed = Rep(directed)

        rep = rep.Copy()

        primary = rep.Get(left, right, itype, directed)
        attributes = rep.Without(primary)
        attributes = attributes.Level(1)
        
        #package data if needed
        primary = primary.Level(1).To(0,1,2,Do=_.Cast("bytes"))
        
        #some checks
        assert len(primary.Slices) == 4, "Edge identifiers should consist of four slices"
        dims = set([pslice.dims for pslice in primary.Slices]) | set([aslice.dims for aslice in attributes.Slices])
        assert len(dims) == 1, "Multiple root dimensions not allowed"

        if addNodes:
            nodes = (primary.Get(0).Set() | primary.Get(1).Set()).Elem()
            self.AddNodes(nodes)

        edgeids = primary.Array(0).ToPython()
        cedgeids = self._server.createEdges(self._networkid, edgeids[0], edgeids[1], edgeids[2], edgeids[3], False)
        
        for pos, slice in enumerate(attributes.Slices):
            attribute = attributes.Get(pos)
            xtype, attribute = self._prepareAttribute(attribute)
            attrmap = dict(zip(cedgeids, attribute.ToPython()))
            self._server.addEdgeAttributes(slice.name, xtype, attrmap)


    def _prepareAttribute(self, attribute):
        assert len(attribute.Slices) == 1, "Attribtue slice nr should be 1"
        slice = attribute.Slices[0]
        if(isinstance(slice.type, rtypes.TypeBool) and not slice.type.has_missing):
            xtype = "BOOLEAN"
        elif(isinstance(slice.type, rtypes.TypeInteger) and not slice.type.has_missing):
            xtype = "INTEGER"
        elif(isinstance(slice.type, rtypes.TypeReal64) and not slice.type.has_missing):
            xtype = "FLOATING"
            attribute = attribute.Each(nan_to_zero,dtype="real64")
        else:
            xtype = "STRING"
            attribute = attribute.Cast("bytes")
        
        return (xtype, attribute)

    def GetLayoutNames(self):
        return Rep(self._server.getLayoutNames())

    def PerformLayout(self, name=None):
        if name is None:
            self._server.performDefaultLayout()
        else:
            self._server.performLayout(self._networkid,name)

    def SetNodeLabel(self, name=None):
        if name is None:
            name = "canonicalName"
        self._server.setNodeLabel(self._networkid, name, "", "default")

    def Nodes(self): 
        return Rep(self._server.getNodes(self._networkid),dtype="[nodes:*]<bytes")

    def Edges(self):
        return Rep([parse(edgeid) for edgeid in self._server.getEdges(self._networkid)],dtype="[edges:*]<(edgeid=bytes, left=bytes,right=bytes,type=bytes)")
    
    def SelNodes(self): 
        return Rep(self._server.getSelectedNodes(self._networkid),dtype="[selnodes:*]<bytes")

    def SelEdges(self):
        return Rep([parse(edgeid) for edgeid in self._server.getSelectedEdges(self._networkid)],dtype="[seledges:*]<(edgeid=bytes, left=bytes,right=bytes,type=bytes)")

    def NodeAttributeInfo(self):
        attr_names = self._server.getNodeAttributeNames()
        if 'hiddenLabel' in attr_names:
            del attr_names[attr_names.index('hiddenLabel')]
        attr_types = [self._server.getNodeAttributeType(attr_name) for attr_name in attr_names]
        return Rep((attr_names, attr_types), dtype="(name=[node_attributes:*]<bytes,type=[node_attributes:*]<bytes)")

    def EdgeAttributeInfo(self):
        attr_names = self._server.getEdgeAttributeNames()
        attr_types = [self._server.getEdgeAttributeType(attr_name) for attr_name in attr_names]
        return Rep((attr_names, attr_types), dtype="(name=[edge_attributes:*]<bytes,type=[edge_attributes:*]<bytes)")

    def NodeAttributes(self, nodes=None, attributes=None):
        if(isinstance(nodes,representor.Representor)):
            nodes = nodes.ToPython()
        if(isinstance(attributes, representor.Representor)):
            attributes = attributes.ToPython()

        if nodes is None:
            nodes = self._server.getNodes(self._networkid)      

        xattributes = self.NodeAttributeInfo()
        if attributes is None:
            attributes = xattributes
        else:   
            attributes = xattributes[_.name |In| attributes]

        columns = []
        fieldnames = []
        fieldtypes = []
        dimpath=dimpaths.DimPath(dimensions.Dim(shape=UNDEFINED,has_missing=False,dependent=(),name="node_attributes"))
        for attribute, atype in attributes.Tuple()():
            atype = self._prepareType(atype)
            has_attribute = util.darray(self._server.nodesHaveAttribute(attribute,nodes) ,bool)
            if False in has_attribute:
                atype.has_missing=True
                res = util.darray([Missing] * len(nodes))
                fnodes = list(util.darray(nodes)[has_attribute])
                res[has_attribute] = self._server.getNodesAttributes(attribute,fnodes)
            else:
                res = self._server.getNodesAttributes(attribute, nodes)
            columns.append(res)
            fieldnames.append(util.valid_name(attribute))
            fieldtypes.append(rtypes.TypeArray(dims=dimpath,subtypes=(atype,)))
            
        dtype = rtypes.TypeTuple(fieldnames=tuple(fieldnames),subtypes=tuple(fieldtypes))
        return Rep(columns,dtype=dtype)

    def EdgeAttributes(self, edges=None, attributes=None):
        if(isinstance(edges,representor.Representor)):
            edges = edges.ToPython()
        if(isinstance(attributes, representor.Representor)):
            attributes = attributes.ToPython()

        if edges is None:
            edges = self._server.getEdges(self._networkid)      

        xattributes = self.EdgeAttributeInfo()
        if attributes is None:
            attributes = xattributes
        else:   
            attributes = xattributes[_.name |In| attributes]

        columns = []
        fieldnames = []
        fieldtypes = []
        dimpath=dimpaths.DimPath(dimensions.Dim(shape=UNDEFINED,has_missing=False,dependent=(),name="edge_attributes"))
        for attribute, atype in attributes.Tuple()():
            atype = self._prepareType(atype)
            has_attribute = util.darray(self._server.edgesHaveAttribute(attribute,edges) ,bool)
            if False in has_attribute:
                atype.has_missing=True
                res = util.darray([Missing] * len(edges))
                fedges = list(util.darray(edges)[has_attribute])
                res[has_attribute] = self._server.getEdgesAttributes(attribute,fedges)
            else:
                res = self._server.getEdgesAttributes(attribute, edges)
            columns.append(res)
            fieldnames.append(util.valid_name(attribute))
            fieldtypes.append(rtypes.TypeArray(dims=dimpath,subtypes=(atype,)))
            
        dtype = rtypes.TypeTuple(fieldnames=tuple(fieldnames),subtypes=tuple(fieldtypes))
        return Rep(columns,dtype=dtype)


    def _prepareType(self, cytype):
        if(cytype == "FLOATING"):
            return rtypes.TypeReal64()
        elif(cytype == "INTEGER"):
            return rtypes.TypePlatformInt()
        elif(cytype == "BOOLEAN"):
            return rtypes.TypeBoolean()
        elif(cytype == "STRING"):
            dim = dimensions.Dim(shape=UNDEFINED, has_missing=False, dependent=(True,))
            return rtypes.TypeBytes(dims=dimpaths.DimPath(dim))
        else:
            return rtypes.unknown

    def _getNodeCount(self):
        return self._server.countNodes(self._networkid)
    NodeCount=property(fget=_getNodeCount)

    def _getEdgeCount(self):
        return self._server.countEdges(self._networkid)
    EdgeCount=property(fget=_getEdgeCount)

    def _getName(self):
        return self._server.getNetworkTitle(self._networkid)
    Name = property(fget=_getName)

    def __repr__(self):
        return "<CyNetwork " + str(self.Name) + ": " + str(self.NodeCount) + " nodes, " + str(self.EdgeCount) + " edges>"
      
    def CreateDiscreteColorMap(self, elems, colormap=None):
        import matplotlib.cm
        import matplotlib.colors
        if colormap is None:
            colormap = matplotlib.cm.gist_rainbow
        cconv = matplotlib.colors.ColorConverter()
        z = set(elems)
        res = {}

        steps = numpy.linspace(0.0, 1.0, len(z))
        colors = colormap(steps)
        for elem, colorrow in zip(z, colors):
            res[elem] = matplotlib.colors.rgb2hex(cconv.to_rgb(colorrow))

        return res

def parse(edgeid):
    split = edgeid.split(' (')
    left = split[0]
    rest = " (".join(split[1:])
    split = rest.split(') ')
    right = split[-1]
    type = " )".join(split[:-1])
    return (edgeid, left, right, type)




import urllib2
import urlparse
import os
from os import path
from distutils.dir_util import mkpath

import tarfile
import zipfile
import gzip
import bz2

class DownloadCache(object):
    def __init__(self, folder=None):
        if(folder is None):
            folder = path.expanduser("~") + "/.ibidas/data"
        self.set_download_folder(folder)

    def url_to_file(self, url):
        return url.split('/')[-1]

    def __call__(self, url, dbname=None, filename=None):
        if(filename is None):
            filename = self.url_to_file(url)
        if(dbname is None):
            dbname = urlparse.urlparse(url).netloc

        file_path = self.get_filename(dbname, filename) 
        if not path.exists(file_path):
            self.download(url, file_path)
        
        return file_path

    def set_download_folder(self, folder):
        self.folder = folder

    def check_dir(self, dir):
        if not path.isdir(dir):
            assert not path.exists(dir), "Folder: " + str(dir) + " already exists as non-directory"
            mkpath(dir)

    def get_filename(self, dbname, filename):
        self.check_dir(self.folder)
        if(dbname):
            self.check_dir(self.folder + "/" + dbname)
            return self.folder + "/" + dbname + "/" + filename
        else:
            return self.folder + "/" + filename

    def download(self, url, file_name):
        u = urllib2.urlopen(url)
        f = open(file_name, 'w')
        meta = u.info()
        file_size = int(meta.getheaders("Content-Length")[0])
        print "Downloading: %s Bytes: %s" % (file_name, file_size)

        file_size_dl = 0
        block_sz = 8192
        while True:
            buffer = u.read(block_sz)
            if not buffer:
                break

            file_size_dl += block_sz
            f.write(buffer)
            status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
            status = status + chr(8)*(len(status)+1)
            print status,

        f.close()

def is_gzipfile(filename):
    try:
        gzip.open(filename,'r')
        gzip.close()
        return True
    except Exception:
        return False


def Unpack(filename, subfiles=None, force=False):
    p = path.dirname(filename)
    unpackpath = path.join(p, '/' + filename + '_unpacked/')
    if not path.isdir(unpackpath):
        os.mkdir(unpackpath)
    osubfiles = subfiles
    if isinstance(subfiles, str):
        subfiles = [subfiles]
    res = []
    statinfo = os.stat(filename)
    if zipfile.is_zipfile(filename):
        z = zipfile.ZipFile(filename,'r')
        if(not subfiles):
            subfiles = z.namelist()
        for subfile in subfiles:
            nspath = path.join(unpackpath, subfile)
            if not path.exists(nspath):
                z.extract(subfile, unpackpath)
            elif statinfo.st_ctime > os.stat(nspath).st_ctime:
                assert force, "Unpacking would overwrite older version. Delete the old data (in " + str(unpackpath) + ") or set force=True to proceed"
                os.unlink(nspath)
            res.append(nspath)

    elif tarfile.is_tarfile(filename):
        z = tarfile.open(filename, 'r')
        if(not subfiles):
            subfiles = z.getnames()
        for subfile in subfiles:
            nspath = path.join(unpackpath, subfile)
            if not path.exists(nspath):
                z.extract(subfile, unpackpath)
            elif statinfo.st_ctime > os.stat(nspath).st_ctime:
                assert force, "Unpacking would overwrite older version. Delete the old data (in " + str(unpackpath) + ") or set force=True to proceed"
                os.unlink(nspath)
                z.extract(subfile, unpackpath)
            res.append(nspath)
    else:
        return filename

    if isinstance(osubfiles, str):
        return res[0]
    else:
        return res        
        
            
            

    


    



from collections import defaultdict, Iterable
import operator
import numpy

import rtypes

_delay_import_(globals(),"dimensions")
_delay_import_(globals(),"casts")
_delay_import_(globals(),"..utils","util","cutils","sparse_arrays")
_delay_import_(globals(),"..utils.missing","Missing")


in1_type_ops = defaultdict(set)
in2_type_ops = defaultdict(set)
out_type_ops = defaultdict(set)
op_type_ops = defaultdict(set)

class Operation(object):
    __slots__ = ["check_func", "exec_funcs"]
    def __init__(self, check_func, **exec_funcs):
        self.check_func = check_func
        self.exec_funcs = exec_funcs

def addOps(in1_type_cls, in2_type_cls, ops, operation):
    """Add ops
        in1_type_cls: in type class, can also be sequence
        in2_type_cls: in type class, can also be sequence
        ops: python string representation of operation (e.g. __add__)
             can also be sequence
        operation: an Operation object
    """
    if(isinstance(in1_type_cls, Iterable)):
        for in1cls in in1_type_cls:
            in1_type_ops[in1cls].add(operation)
    else:
        in1_type_ops[in1_type_cls].add(operation)
    
    if(isinstance(in2_type_cls, Iterable)):
        for in2cls in in2_type_cls:
            in2_type_ops[in2cls].add(operation)
    else:
        in2_type_ops[in2_type_cls].add(operation)

    if(isinstance(ops, Iterable)
                        and not isinstance(ops, str)):
        for op in ops:
            op_type_ops[op].add(operation)
    else:
        op_type_ops[ops].add(operation)


def unop_type(intype, op):
    operations = op_type_ops[op] & in1_type_ops[intype.__class__]

    for oper in operations:
        res = oper.check_func(intype, op)
        if(not res is False):
            return (res, oper) 
    raise RuntimeError, "Could not find a valid operation handler for " + \
            op + " on " + str(intype) 
        

def binop_type(ltype, rtype, op, outtype=None):
    operations = op_type_ops[op] & in1_type_ops[ltype.__class__] & \
                                    in2_type_ops[rtype.__class__]    
    for oper in operations:
        res = oper.check_func(ltype, rtype, op)
        if(not res is False):
            return (res, oper) 

    raise RuntimeError, "Could not find a valid operation handler for " + \
            str(ltype) + " " + op + " " + str(rtype)
        

def check_arith(in1_type, in2_type, op):#{{{
    in1_impli_cls = casts.findImplicitCastTypes(in1_type.__class__)
    in2_impli_cls = casts.findImplicitCastTypes(in2_type.__class__)
    out_impli_cls = in1_impli_cls & in2_impli_cls
    
    out_impli_cls = [ocls for ocls in out_impli_cls \
                            if issubclass(ocls,rtypes.TypeNumber)]
    if(not out_impli_cls):
        return False
  
    out_impli_cls = rtypes.mostSpecializedTypesCls(out_impli_cls)
    assert len(out_impli_cls) == 1, "Multiple output types for " + \
                  "arithmetic operation found: " + str(out_impli_cls)
    out_cls = out_impli_cls[0]
    
    out_type = out_cls(in1_type.has_missing or in2_type.has_missing)
    return out_type#}}}

numpy_cmp = {'__eq__':numpy.equal,#{{{
            '__ne__':numpy.not_equal,
            '__le__':numpy.less_equal,
            '__ge__':numpy.greater_equal,
            '__lt__':numpy.less,
            '__gt__':numpy.greater}

numpy_arith = { '__add__':numpy.add,
                '__radd__':numpy.add,
                '__sub__':numpy.subtract,
                '__rsub__':numpy.subtract,
                '__mul__':numpy.multiply,
                '__rmul__':numpy.multiply,
                '__mod__':numpy.mod,
                '__rmod__':numpy.mod,
                '__div__':numpy.divide,
                '__rdiv__':numpy.divide,
                '__and__':numpy.bitwise_and,
                '__rand__':numpy.bitwise_and,
                '__or__':numpy.bitwise_or,
                '__ror__':numpy.bitwise_or,
                '__xor__':numpy.bitwise_xor,
                '__rxor__':numpy.bitwise_xor,
                }

numpy_unary_arith = {
    "__invert__":numpy.invert,
    "__neg__":numpy.negative,
    "__abs__":numpy.abs
    }

reverse_op = {'__eq__':'__eq__',
            '__ne__':'__ne__',
            '__le__':'__ge__',
            '__ge__':'__le__',
            '__lt__':'__gt__',
            '__gt__':'__lt__',
            '__add__':'__radd__',
            '__radd__':'__add__',
            '__sub__':'__rsub__',
            '__rsub__':'__sub__',
            '__mul__':'__rmul__',
            '__rmul__':'__mul__',
            '__mod__':'__rmod__',
            '__rmod__':'__mod__',
            '__div__':'__rdiv__',
            '__rdiv__':'__div__',
            '__and__':'__rand__',
            '__rand__':'__and__',
            '__or__':'__ror__',
            '__ror__':'__or__',
            '__xor__':'__rxor__',
            '__rxor__':'__xor__',
            }#}}}

def exec_arith(data, type1, type2, typeo, op):
    data1,data2 = data
    if(data1 is Missing or data2 is Missing):
        return Missing
    if util.numpy16up:
        return numpy_arith[op](data1, data2, dtype=typeo.toNumpy())
    else:
        return numpy_arith[op](data1, data2, sig=typeo.toNumpy())

arith_ops = set(numpy_arith.keys())
addOps(rtypes.TypeNumbers, rtypes.TypeNumbers, arith_ops, Operation(check_arith, py=exec_arith))


def check_cmp(in1_type, in2_type, op):#{{{
    out_cls = rtypes.TypeBool
    out_type = out_cls(in1_type.has_missing or in2_type.has_missing)
    return out_type#}}}

def exec_cmp(data, type1, type2, typeo, op):
    #a numpy bug gives all true arrays when using
    #bool as outtype in comparison
    assert isinstance(typeo,rtypes.TypeBool),"Comparison should have result type bool"
    return numpy_cmp[op](data[0], data[1])


cmp_ops = set(numpy_cmp.keys())
addOps(rtypes.TypeNumbers, rtypes.TypeNumbers, cmp_ops, Operation(check_cmp, py=exec_cmp))

#CHECKED UP TO HERE


def exec_cmpgeneral(data1, data2, type1, type2, typeo, op):
    assert isinstance(typeo,rtypes.TypeBool),"Comparison should have result type bool"
    
    #a numpy bug gives all true arrays when using
    #bool as outtype in comparison
    res = getattr(data1, op)(data2)
    if(res is NotImplemented):
        res = getattr(data2, reverse_op[op])(data1)
    if(res is NotImplemented):
        raise RuntimeError, "Not implemented error in exec_cmpgeneral"
    return res

addOps(rtypes.TypeStrings, rtypes.TypeStrings, cmp_ops, Operation(check_cmp, py=exec_cmpgeneral))

def check_cmpset(in1_type, in2_type, op):#{{{
    
    if(isinstance(in1_type, rtypes.TypeSet) and 
                    isinstance(in2_type, rtypes.TypeSet)):
        pass
    elif isinstance(in1_type, rtypes.TypeSet) and op in set_funcs:
        pass
    elif isinstance(in2_type, rtypes.TypeSet) and op in set_funcs:
        pass
    else:    
        return False
    return check_cmp(in1_type, in2_type, op)#}}}
    

set_funcs = {'__or__':"union", "__ror__":"union", "__and__":"intersection", 
        "__rand__":"intersection",  "__sub__":"difference", 
        "__rsub__":"difference", "__xor__":"symmetric_difference",
        "__rxor__":"symmetric_difference",
        "__le__":"issubset", "__ge__":"issuperset"}


def exec_object_func(data1, data2, type1, type2, otype, op):
    funcget = operator.attrgetter(op)
    if(isinstance(type1, rtypes.TypeAny) and type1.has_missing or
       isinstance(type2, rtypes.TypeAny) and type2.has_missing):
        def func(x, y):
            if(x is Missing or y is Missing):
                return Missing
            else:
                return funcget(x)(y)
    else:  
        func = lambda x, y: funcget(x)(y)
    if(isinstance(data1, numpy.ndarray) and 
                                  not type1.__class__ is rtypes.TypeArray):
        if(isinstance(data2, numpy.ndarray) and 
                                      not type2.__class__ is rtypes.TypeArray):
            res = util.darray([func(left, right) 
                                 for left, right in zip(data1, data2)], 
                                 otype.toNumpy()).view(sparse_arrays.FullSparse)
        else:
            res = util.darray([func(left, data2) 
                                 for left in data1], 
                                 otype.toNumpy()).view(sparse_arrays.FullSparse)
    else:
        if(isinstance(data2, numpy.ndarray) and 
                                      not type2.__class__ is rtypes.TypeArray):
            res = util.darray([func(data1, right) 
                                 for right in data2], 
                                 otype.toNumpy()).view(sparse_arrays.FullSparse)
        else:
            res = func(data1, data2)
            
            if(otype):
                res = otype.toNumpy().type(res)
    return res   

def exec_arrayarray(data1, data2, type1, type2, typeo, op):
    #a numpy bug gives all true arrays when using
    #bool as outtype in comparison
    if(op == "__add__"):
        return numpy.concatenate((data1, data2)).view(sparse_arrays.FullSparse)
    elif(op == "__radd__"):
        return numpy.concatenate((data2, data1)).view(sparse_arrays.FullSparse)
    else:
        raise RuntimeError, "Unrecognized operation " + op

def check_arrayarray(in1_type, in2_type, op):#{{{
    
    if(not (isinstance(in1_type, rtypes.TypeArray) and 
                    isinstance(in2_type, rtypes.TypeArray))):
        return False

    in1_subtype = in1_type.subtypes[0]
    in2_subtype = in2_type.subtypes[0]
    subtype = casts.castImplicitCommonType(in1_subtype, in2_subtype)
    
    odim1 = in1_type.dims[0]
    odim2 = in2_type.dims[0]

    #FIX variable
    ndim = dimensions.Dim(UNDEFINED, 
                          odim1.variable or odim2.variable,  
                          odim1.has_missing or odim2.has_missing)
    return rtypes.TypeArray(in1_type.has_missing or in2_type.has_missing, 
                          (ndim,), (subtype,))
    #}}}
array_ops = set(["__add__", "__radd__"])
addOps(rtypes.TypeArrays, rtypes.TypeArrays, 
       array_ops, Operation(check_arrayarray, py=exec_arrayarray))



def exec_cmpset(data1, data2, type1, type2, typeo, op):
    #a numpy bug gives all true arrays when using
    #bool as outtype in comparison
    rop = reverse_op[op]
    if(op in set_funcs):
        func = set_funcs[op]
        rfunc = set_funcs[rop]
    else:
        func = op
        rfunc = rop
    if(isinstance(type1, rtypes.TypeSet)):
        return exec_object_func(data1, data2, type1, type2, typeo, func)
    else:
        return exec_object_func(data2, data1, type2, type1, typeo, rfunc)



def check_setset(in1_type, in2_type, op, out_type):#{{{
    
    if(isinstance(in1_type, rtypes.TypeSet) and 
                    isinstance(in2_type, rtypes.TypeSet)):
        pass
    elif(isinstance(in1_type, rtypes.TypeSet) and 
                                    (op in set_funcs and op != "__rsub__")):
        pass
    elif(isinstance(in2_type, rtypes.TypeSet) and 
                                    (op in set_funcs and op != "__sub__")):
        pass
    else:    
        return False

    in1_subtype = in1_type.subtypes[0]
    in2_subtype = in2_type.subtypes[0]
    subtype = casts.castImplicitCommonType(in1_subtype, in2_subtype)
    
    odim1 = in1_type.dims[0]
    odim2 = in2_type.dims[0]
    ndim = dimensions.Dim(UNDEFINED, 
                          odim1.variable or odim2.variable, 
                          odim1.has_missing or odim2.has_missing)
    return rtypes.TypeSet(in1_type.has_missing or in2_type.has_missing, 
                          (ndim,), (subtype,))
    #}}}

def exec_setset(data1, data2, type1, type2, typeo, op):
    #a numpy bug gives all true arrays when using
    #bool as outtype in comparison
    rop = reverse_op[op]
    func = set_funcs[op]
    rfunc = set_funcs[rop]

    if(isinstance(type1, rtypes.TypeSet)):
        return exec_object_func(data1, data2, type1, type2, typeo, func)
    else:
        return exec_object_func(data2, data1, type2, type1, typeo, func)

setchange_ops = set(["__or__", "__ror__", "__and__", "__rand__", "__sub__", 
                                        "__rsub__", "__xor__", "__rxor__"])


addOps(rtypes.TypeArrays, rtypes.TypeArrays, 
       cmp_ops, Operation(check_cmpset, py=exec_cmpset))
addOps(rtypes.TypeArrays, rtypes.TypeArrays, 
       setchange_ops, Operation(check_setset, py=exec_setset))


def check_any(in1_type, in2_type, op):#{{{
    if(in1_type.__class__ is rtypes.TypeUnknown or 
       in2_type.__class__ is rtypes.TypeUnknown):
        return rtypes.unknown
    else:
        return rtypes.TypeAny(in1_type.has_missing or in2_type.has_missing)
#}}}

any_types = set([rtypes.TypeUnknown, rtypes.TypeAny])
all_ops = set(reverse_op.keys())

addOps(any_types, any_types, all_ops, 
       Operation(check_any, py=exec_object_func))

def check_unary_any(in_type, op):
    return in_type

def exec_object_unaryfunc(data1, type1, otype, func):
    funcget = operator.attrgetter(func)
    if(isinstance(type1, rtypes.TypeAny) and type1.has_missing):
        def func(x):
            if(x is Missing):
                return x
            else:
                return funcget(x)()
    else:  
        func = lambda x: funcget(x)()

    if(isinstance(data1, numpy.ndarray) and 
                                    not type1.__class__ is rtypes.TypeArray):
        func = numpy.vectorize(func)
        res = numpy.cast[otype.toNumpy()](func(data1)).view(sparse_arrays.FullSparse)
    else:
        res = func(data1)
            
        if(otype):
            res = otype.toNumpy().type(res)
    return res   


unary_ops = set(numpy_unary_arith.keys())
#addOps(rtypes.TypeNumbers, (), rtypes.TypeNumbers, unary_ops, 
#       Operation(check_unary_any, py=exec_object_unaryfunc))

def check_unary_arith(in_type, op):#{{{
    in_impli_cls = casts.findImplicitCastTypes(in_type.__class__)
    out_impli_cls = in_impli_cls
    
    out_impli_cls = [ocls for ocls in out_impli_cls \
                            if issubclass(ocls,rtypes.TypeNumber)]
    if(not out_impli_cls):
        return False
  
    out_impli_cls = rtypes.mostSpecializedTypesCls(out_impli_cls)
    assert len(out_impli_cls) == 1, "Multiple output types for " + \
    "arithmetic operation found: " + str(out_impli_cls)
    out_cls = out_impli_cls[0]
    
    return out_cls(in_type.has_missing)#}}}


def exec_unary_arith(data1, intype, outtype, op):
    if util.numpy16up:
        return numpy_unary_arith[op](data1, dtype=outtype.toNumpy())
    else:
        return numpy_unary_arith[op](data1, sig=outtype.toNumpy())

addOps(rtypes.TypeNumbers, (), unary_ops, 
        Operation(check_unary_arith, py=exec_unary_arith))






"""Engines used to process data."""
import passes
from wrappers import python

select_engine = passes.PassManager(log=False)
select_engine.register(passes.EnsureInfo)
select_engine.register(passes.CreateGraph)
select_engine.register(passes.AnnotateRepLinks)
#select_engine.register(passes.DebugVisualizer)
select_engine.register(passes.PrePeepHole)
select_engine.register(passes.CreateExpressions)
select_engine.register(passes.WrapperPlanner,debug_mode=False)
select_engine.register(passes.PythonPeepHole)
#select_engine.register(passes.DebugVisualizer)
select_engine.register(passes.Transposer)
#select_engine.register(passes.DebugVisualizer)
select_engine.register(passes.SerializeExec)
select_engine.register(python.PyExec,debug_mode=False)
#select_engine.register(passes.DebugVisualizer)


debug_engine = passes.PassManager(log=True)
debug_engine.register(passes.EnsureInfo)
debug_engine.register(passes.CreateGraph)
debug_engine.register(passes.AnnotateRepLinks)
debug_engine.register(passes.DebugVisualizer)
debug_engine.register(passes.PrePeepHole)
debug_engine.register(passes.CreateExpressions)
debug_engine.register(passes.WrapperPlanner,debug_mode=True)
debug_engine.register(passes.PythonPeepHole)
#debug_engine.register(passes.DebugVisualizer)
debug_engine.register(passes.Transposer)
#debug_engine.register(passes.DebugVisualizer)
debug_engine.register(passes.SerializeExec)
debug_engine.register(python.PyExec,debug_mode=True)
debug_engine.register(passes.DebugVisualizer)




import manager
import create_graph, prepeephole, annotate_replinks
from ..utils import util
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, F_CACHE
from ..constants import *
from annot_objs import *



class CreateExpressions(VisitorFactory(prefixes=("distribute","createExp"), 
                                      flags=NF_ERROR | F_CACHE), manager.Pass):
    after=set([create_graph.CreateGraph, annotate_replinks.AnnotateRepLinks, prepeephole.PrePeepHole])

    @classmethod
    def run(cls, query, run_manager):
        self = cls()
        query_root = query.root
        self.graph = run_manager.pass_results[create_graph.CreateGraph]
        self.links = self.graph.na['links']

        self.expressions = {}
        self.distribute(self.graph.root)
        self.graph.checkGraph()
        return self.expressions

    def distributeOp(self, node, lastlink=None):
        if(node in self.links):
            selflink = self.links[node]
        else:
            selflink = lastlink
        
        if(not id(selflink) in self.expressions):
            e = self.createExp(selflink)
            self.expressions[id(selflink)] = e
        else:
            e = self.expressions[id(selflink)]

        e.addAllSlice(node)
        if(not selflink is lastlink):
            e.addOutSlice(node)
      
        for edge in self.graph.edge_target[node]:
            sourcelink = self.distribute(edge.source, selflink)
            if not sourcelink is selflink:
                e.addInSlice(edge.target)
                
        self.links[node] = e
        return selflink
        

    def createExpRepresentor(self, rep):
       return Expression(rep.__class__.__name__, rep)

    def createExpFilter(self, rep):
       return FilterExpression(rep.__class__.__name__, rep)

    def createExpBinaryFuncElemOp(self, rep):
        return BinFuncElemExpression(rep.__class__.__name__, rep)
    
    def createExpMatch(self, rep):
        return MatchExpression(rep.__class__.__name__, rep)



import numpy
import hashlib
from itertools import izip
import operator
import collections
try:
    from IPython.Debugger import Tracer; debug_here = Tracer()
except ImportError:
    try:
        from IPython.core.debugger import Tracer; debug_here = Tracer()
    except:
        pass
import cPickle, zlib
import math
import os.path
import gc, sys
import itertools
import csv
import cutils
from logging import error,warning,info,debug

#objs: objs to find names for
#req_objs: objs which should be in same frame
def find_names(objs, req_objs=[],skip_frames=True):
    #after: http://pythonic.pocoo.org/2009/5/30/finding-objects-names
    result = []
    referrers = dict()
    refsetobjs = []

    frame = sys._getframe()
    #refs should be in locals dict
    frameobj = []
    for fcounter,frame in enumerate(iter(lambda: frame.f_back, None)):
        fid = id(frame.f_locals)
        if not skip_frames or fcounter >= 1:
            referrers[fid] = frame.f_locals
            frameobj.append(fid)
    refsetobjs.append(set(frameobj))

    for obj in itertools.chain(objs,req_objs):
        refsetobj = [id(referrer) for referrer in gc.get_referrers(obj) if isinstance(referrer, dict)]
        refsetobjs.append(set(refsetobj))
    refs_with_all_objs = reduce(operator.__and__,refsetobjs)

    for refid in frameobj:
        if not refid in refs_with_all_objs:
            continue
        referrer = referrers[refid]
        result = []
        for obj in objs:
            for k, v in referrer.iteritems():
                if v is obj and not k.startswith('_') and not k[:2].isupper():
                    result.append(k)
        if(len(result) == len(objs)):
            return result
    return []

_delay_import_(globals(),"missing","Missing")
def save_rep(r, filename):
    f = open(filename, 'wb')
    s = cPickle.dumps(r,protocol=2)
    s = zlib.compress(s)
    f.write(s)


def save_csv(r, filename):
    f = open(filename,'wb')
    r= r.Array(tolevel=1)
    data = r.Cast(str).Tuple().ToPython();
    if filename.endswith('tsv'):
        w = csv.writer(f,delimiter='\t');
    else:
        w = csv.writer(f);
    w.writerow(r.Names);
    w.writerows(data);

def load_rep(filename):
    f = open(filename, 'rb')
    s = zlib.decompress(f.read())
    return cPickle.loads(s)


def valid_name(name):
    name = name.lower()
    newname = []
    for char in name:
        if(char.isalnum() or char == "_"):
            newname.append(char)
        else:
            newname.append("_")
    newname = "".join(newname)
    newname = newname.strip("_")
    if(not newname):
        return "funknown"
    if(newname[0].isdigit()):
        newname = "_" + newname
    return str(newname)
        

def seqgen(start=0):
    """Generator, generating monotonically increasing indexes"""
    while True:
        start += 1
        yield start


def unique(seq):
    """Given a list or tuple, returns a list or tuple in which each element
    occurs only once, while keeping the original order or the list"""
    seen_elems = set()
    res = [elem for elem in seq 
        if not elem in seen_elems and 
        not seen_elems.add(elem)]

    if(isinstance(seq, tuple)):
        return tuple(res)
    else:
        return res

def select(seq, index):
    """Given list or tuple in `seq`, select using `index`, 
    and returns list or tuple.

    :param seq: list or tuple
    :param index: int, long, sequence of indexes, sequence of booleans
    """
    if(isinstance(index, slice)):
        return seq[index]
    elif(operator.isSequenceType(index)) :
        if(len(index) == 0):
            return seq.__class__()

        if(isinstance(index[0], bool)):
            res = [elem for elem, indicator in zip(seq, index) if indicator]
        else:
            res = [seq[idx] for idx in index]
        
        if(isinstance(seq, tuple)):
            return tuple(res)
        return res

    if(isinstance(seq, tuple)):
        return (seq[index],)
    else:
        return [seq[index]]


def index_field(seq, fieldname):
    """Given a sequence in `seq`, retrieves attribute 
    with name `fieldname` from all
    elements, and creates a dictionary from fieldname value to element.
    
    :param seq: sequence of objects
    :param fieldname: attribute available in each object
    """
    func = operator.attrgetter(fieldname)
    return dict((func(elem), elem) for elem in seq)


def replace_in_tuple(tpl, replace):
    """Replaces objects in `tpl`, by using dictionary in `replace`"""
    return tuple([replace.get(elem, elem) for elem in tpl])

def delete_from_tuple(tpl, remove):
    return tuple([elem for elem in tpl if not elem in remove])


def zip_broadcast(*elems):
    lengths = set([len(elem) for elem in elems])
    lengths.discard(1)
    assert len(lengths) <= 1, "Number of elements in zip broadcast should be equal (or 1)"
    if(lengths):
        length = lengths.pop()
    else:
        length = 1
  
    elems = list(elems)
    for pos, elem in enumerate(elems):
        if(len(elem) == 1):
            elems[pos] = elem * length
    return zip(*elems)

def zip_broadcast_old(*elems):
    iters = []
    firstres = []
    for elem in elems:
        if(isinstance(elem, collections.Iterable)):
            iters.append(iter(elem))
            firstres.append(iters[-1].next())
        else:
            iters.append(None)
            firstres.append(elem)

    yield firstres

    secondres = []
    for pos, elem in enumerate(iters):
        if(elem is None):
            secondres.append(firstres[pos])
        else:
            try:
                secondres.append(elem.next())        
            except StopIteration:
                secondres.append(firstres[pos])
                iters[pos] = None

    if(all(elem is None for elem in iters)):
        raise StopIteration

    yield tuple(secondres)

    idxs = range(len(iters))
    while True: 
        res = [None] * len(iters)
        for pos in idxs:
            elem = iters[pos]
            if(not elem):
                res[pos] = firstres[pos]
            else:
                res[pos] = elem.next()
        yield res

def filter_missing(func):
    def fmissing(x, *params):
        if(x is Missing):
            return x
        else:
            return func(x, *params)
    return fmissing

def create_strtable(table):
    column_width = []
    for col in table:
        col_width = max([len(row) for row in col])
        column_width.append(col_width)

    res = ""
    for idx in range(len(table[0])):
        for cwidth, col in zip(column_width, table):
            res += col[idx].ljust(cwidth + 1)
        res += "\n"
    return res

def transpose_table(table):
    ncols = len(table[0])
    cols = []
    for col in range(ncols):
        cols.append([row[col] for row in table])
    return cols
        
def contained_in(inner_tuple, outer_tuple):
    pos = 0
    linner_tuple = len(inner_tuple)
    louter_tuple = len(outer_tuple)

    if(linner_tuple > louter_tuple):
        return False
    if(not inner_tuple):
        return (0,0)

    try:
        first_pos = pos = outer_tuple.index(inner_tuple[0])
    except ValueError, e:
        return False

    for elem in inner_tuple[1:]:
        try:
            pos = outer_tuple.index(elem, pos)
        except ValueError, e:
            return False

    return (first_pos, pos + 1)
    

class farray(numpy.ndarray):
    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                   self.shape == other.shape and
             (self.view(numpy.ndarray) == other.view(numpy.ndarray)).all())

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        if(len(self) != len(other)):
            return len(self) > len(other)
        else:
            for x, y in izip(self, other): 
                if(x == y):
                    continue
                return x > y
            else:
                return False

    def __ge__(self, other):
        if(len(self) != len(other)):
            return len(self) >= len(other)
        else:
            for x, y in izip(self, other): 
                if(x == y):
                    continue
                return x >= y
            else:
                return True

    def __lt__(self, other):
        if(len(self) != len(other)):
            return len(self) < len(other)
        else:
            for x, y in izip(self, other): 
                if(x == y):
                    continue
                return x < y
            else:
                return False
        return 

    def __le__(self, other):
        if(len(self) != len(other)):
            return len(self) <= len(other)
        else:
            for x, y in izip(self, other): 
                if(x == y):
                    continue
                return x <= y
            else:
                return True
        return 

    def __hash__(self):
        res = hash(self.shape) ^ hash(self.dtype)
        if(self.dtype == object):
            for elem in self.ravel():
                res ^= hash(elem)
        else:
            res ^= hash(hashlib.md5(self.data).hexdigest())
        return res


def open_file(filename,mode='r'):
    filename = os.path.expanduser(filename)
    if(filename.endswith("gz")):
        import gzip
        file = gzip.open(filename)
    else:
        file = open(filename,mode=mode)
    return file


#after IPython implementation
units = [u"s", u"ms",u'us',"ns"]
scaling = [1, 1e3, 1e6, 1e9]

def format_runtime(rtime):
    if rtime > 0.0 and rtime < 1000.0:
        order = min(-int(math.floor(math.log10(rtime)) // 3), 3)
    elif rtime >= 1000.0:
        order = 0
    else:
        order = 3
    return u"%.*g %s" % (3, rtime * scaling[order],units[order])


def check_module(modulename):
    try:
        __import__(modulename)
        return True
    except ImportError:
        return False

#from: http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
class memoized(object):
   """Decorator that caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned, and
   not re-evaluated.
   """
   def __init__(self, func):
      self.func = func
      self.cache = {}
      self.__name__ = func.__name__
   def __call__(self, *args, **kwargs):
       if kwargs: #uncacheable
            return self.func(*args, **kwargs)
       try:
          return self.cache[args]
       except KeyError:
          value = self.func(*args)
          self.cache[args] = value
          return value
       except TypeError:
          # uncachable -- for instance, passing a list as an argument.
          # Better to not cache than to blow up entirely.
          return self.func(*args)

   def __repr__(self):
       """Return the function's docstring."""
       return self.func.__doc__

   def __get__(self, obj, objtype):
       """Support instance methods."""
       return functools.partial(self.__call__, obj)

def check_realseq(data):
    return isinstance(data,collections.Sequence) and not isinstance(data, basestring)

def get_shape(data, lengths, pos=0):
    if check_realseq(data):
        if len(lengths) > pos + 1 and len(lengths[pos + 1]) <= 1:
            for elem in data:
                if not get_shape(elem,lengths,pos+1):
                    break
        lengths[pos].add(len(data))
    else:
        lengths[pos].add(0)
    return len(lengths[pos]) == 1

def fill(data, x, pos):
    if pos > 1:
        for i in range(x.shape[0]):
            fill(data[i],x[i,...],pos-1)
    else:
        x[:] = data

def replace_darray(data, type=object, maxdim=1, mindim=0):
    if maxdim == 1 and type == object:
        if not check_realseq(data):
            if mindim == 0:  return data
            else: raise ValueError, "Object not deep enough"
        else:
            nlen = len(data)
            x = numpy.zeros((nlen,),dtype=type)
            for pos,e in enumerate(data): #numpy insists on processing the nested data when using x[:] = data, making it really inefficient
                x[pos] = e
            return x
       
    try:
        return cutils.darray(data,type,maxdim,mindim)
    except ValueError:
        if maxdim == 1:
            if not check_realseq(data):
                if mindim == 0:  return data
                else: raise ValueError, "Object not deep enough"
            else:
                nlen = len(data)
                x = numpy.zeros((nlen,),dtype=type)
                x[:] = data
                return x
        #maxdim > 1        
        lengths = [set() for i in range(maxdim)]
        get_shape(data, lengths)
        cusedim = 0
        for lg in lengths:
            if len(lg) == 1:
                cusedim = cusedim + 1
        if cusedim < mindim:
            raise ValueError, "Object not deep enough"
        shape = [lengths[i].pop() for i in range(cusedim)]
        x = numpy.zeros(tuple(shape),dtype=type)
        fill(data,x, len(x.shape))
        return x
        

nversion = numpy.__version__.split('.')    
numpy16up = int(nversion[0]) >= 1 and int(nversion[1]) >= 6

if numpy16up:
    darray = replace_darray
else:
    darray = cutils.darray
    



import time, math
from ..utils import topological_sort, util

class Pass(topological_sort.TopologicalSortable):
    """Pass object. Each pass should inherit from
    this object, and implement a new version of
    the method run."""

    @classmethod
    def run(cls, query, pass_results):
        """Should be reimplemented by child classes. 
        :param query: reference to query context
        :param pass_results: dictionary with results from earlier 
                        passes (accessible by using class as key).

        :rtype: Can return several values:
            - None
            - Object : stored in pass_results
            - (Object, invalidate_passes)
            - (Object, invalidate_passes, add_passes)
        
        """
class PassManager(object):
    """Class to handle passes. Passes are objects which 
    perform some function on a query object."""

    def __init__(self, log=False):
        """Initializes pass manager"""
        self.objs = []
        self.params = []
        self.log = log

    def register(self, objcls, *args, **kwargs):
        """Registers a new pass.
        Passes are ordered according to their topological
        characteristics. If no conflict arises, passes are kept
        in place of their register position. 
        """
        assert issubclass(objcls, Pass), \
            "Only pass classes can be registered"
        self.objs.append(objcls)
        self.params.append((args,kwargs))

    def run(self, query, log=False):
        """Performs registered passes on query, by
        performing a stable topological sort."""

        return PassManagerRun(self, query, self.log or log).run()


class PassManagerRun(object):
    def __init__(self,manager, query, log):
        self.query = query
        self.topoiter = topological_sort.topo_sorted(manager.objs,return_index=True)
        self.pass_results = {}
        self.objects = list(manager.objs)
        self.params = list(manager.params)
        self.log = log


    def invalidate(self, ipass):
        self.topoiter.invalidate(ipass)
        self.pass_results.pop(ipass)

    def add_pass(self, ipass, *params, **kwds):
        self.topoiter.prepend(ipass)
        self.objects.append(ipass)
        self.params.append((params,kwds))

    def run(self):
        if self.log:
            logtable = []
            totalstart = time.time()
        for cur_pass_idx in self.topoiter:
            cur_pass = self.objects[cur_pass_idx]
            param_args, param_kwds = self.params[cur_pass_idx]

            if(self.log):
                start = time.time()
            res = cur_pass.run(self.query, self, *param_args, **param_kwds)
            if(self.log):
                rtime = time.time() - start
                logtable.append((cur_pass.__name__, rtime))
            if(not res is None):
                self.pass_results[cur_pass] = res
                result = res
        if(self.log):
            totaltime = time.time() - totalstart
            restable = []
            for name, rtime in logtable:
                restable.append((name, util.format_runtime(rtime), u"(%.2f%%)" % (100 * (rtime/totaltime)),))
            print util.create_strtable(util.transpose_table(restable))
            print "-------------"
            print u"Total : %s" % (util.format_runtime(totaltime))

        return result



import manager
import ensure_info
import create_graph
from ..utils import util
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, F_CACHE
from ..constants import *
_delay_import_(globals(),"..representor")
_delay_import_(globals(),"..ops")

class WrapperPlanner(VisitorFactory(prefixes=("link","distribute"), 
                                      flags=NF_ERROR | F_CACHE), manager.Pass):
    after=set([create_graph.CreateGraph])

    @classmethod
    def run(cls, query, run_manager, debug_mode):
        self = cls()
        self.graph = run_manager.pass_results[create_graph.CreateGraph]

        passes = set()
        for node in self.graph.nodes:
            if hasattr(node.__class__, "passes"):
                passes.update(node.passes)
        
        for p in passes:
            run_manager.add_pass(p, debug_mode=debug_mode)




import math
import numpy
import operator
import collections
from ..constants import *
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, NF_ELSE

_delay_import_(globals(), "..utils","sparse_arrays","cutils","missing","util")
_delay_import_(globals(), "..utils.missing","Missing")
_delay_import_(globals(), "rtypes")
_delay_import_(globals(), "dimpaths")

class BaseConvertor(object):
    def __init__(self,objectclss=set()):
        self.objectclss = objectclss

    def convert(self,seq,elem_type):
        return seq

class ArrayConvertor(BaseConvertor):
    def convert(self,seq,elem_type):
        """Converts sequence to standard format. Converts None
        to Missing values."""
        return self._convert(dimpaths.getArrayDimPathFromType(elem_type),elem_type,seq)
 
    def _convert(self,dims,elem_type,seq):
        assert isinstance(elem_type,rtypes.TypeArray), "ArrayCOnvertor should be applied to array types"
        
        depth = dims.contigiousFixedNDims()
        if(not depth):# no fixed dims, allow for one non-fixed dim
            if(dims[0].has_missing):
                depth = 1  #no following dims (has_missing)
            else:
                depth = 1 + dims[1:].contigiousFixedNDims()
            variable=True
        else:
            variable=False

        rest_dims = elem_type.dims[depth:]
        if(rest_dims):
            elem_numpy_type = object
            elem_type = rtypes.TypeArray(elem_type.has_missing,dims=rest_dims,subtypes=elem_type.subtypes)
        else:
            elem_numpy_type = elem_type.subtypes[0].toNumpy()

        if(variable):
            seq_numpy_type = object
        else:
            seq_numpy_type = elem_numpy_type
            

        res = []
        for elem in seq:
            if(elem is Missing or elem is None):
                res.append(Missing)
                continue
            #convert to numpy array if needed
            if(not isinstance(elem, numpy.ndarray)):
                if(not isinstance(elem, collections.Sequence)):
                    elem = list(elem)
                elem = util.darray(elem,elem_numpy_type,depth,1)
            
            #assure correct number of dimensions
            if(len(elem.shape) < depth):
                oshape = elem.shape
                rem_ndims = depth - len(elem.shape) + 1
                elem = util.darray(list(elem.ravel()),elem_numpy_type,rem_ndims,rem_ndims)
                elem.shape = oshape + elem.shape[1:]
                assert len(elem.shape) == depth, "Non array values encountered for dims " + str(dims[len(elem.shape):])
            elif(len(elem.shape) > depth):
                oshape = elem.shape
                elem = dimpaths.flatFirstDims(elem,depth-1)
                elem = util.darray([subelem for subelem in elem],object,1,1)
                elem.shape = oshape[:depth]

            if not elem.dtype == elem_numpy_type:
                if(elem_numpy_type.char == 'S' or elem_numpy_type.char == 'U' or elem_numpy_type.char == 'V'):
                    z = numpy.zeros(elem.shape,elem_numpy_type)
                    z[:] = elem
                    elem = z
                else:
                    elem = numpy.cast[elem_numpy_type](elem)

            if(rest_dims):
                elem = self._convert(rest_dims,elem_type,elem)
            res.append(elem)

        if(variable):
            nseq = util.darray(res,seq_numpy_type)
        else:
            nseq = util.darray(res,seq_numpy_type,depth+1,1)

        return nseq

class StringConvertor(BaseConvertor):
    def convert(self,seq,elem_type):
        #deduplicate
        dtype = elem_type.toNumpy()
        if(dtype == object):
            x = seq[:100]
            if(len(x) * 0.6 > len(set(x))):
                r = dict()
                res = []
                for elem in seq:
                    if(elem is Missing or elem is None):
                        res.append(Missing)
                    else:
                        try:
                            elem = r[elem]
                        except KeyError:
                            r[elem] = elem
                        res.append(elem)
                seq = util.darray(res)
        elif(not seq.dtype == dtype):
            #numpy bug: numpy.cast[<string dtype>] always cast to S1, irrespective 
            #requested length
            seq = util.darray(list(seq),numpy.string_)
        return seq


class FloatConvertor(BaseConvertor):
    def convert(self,seq,elem_type):
        if(elem_type.has_missing):
            seqres = []
            for elem in seq:
                if(elem is Missing or elem == ""):
                    seqres.append(Missing)
                else: 
                    seqres.append(float(elem))
        else:
            seqres = [float(elem) for elem in seq]
         
        return util.darray(seqres,elem_type.toNumpy())

class IntegerConvertor(BaseConvertor):
    def convert(self,seq,elem_type):
        if(elem_type.has_missing):
            seqres = []
            for elem in seq:
                if(elem is Missing or elem == ""):
                    seqres.append(Missing)
                else: 
                    seqres.append(int(elem))
        else:
            seqres = [int(elem) for elem in seq]
         
        return util.darray(seqres,elem_type.toNumpy())
 
class SetConvertor(BaseConvertor):

    def convert(self,seq,elem_type):
        if(not self.objectclss or set in self.objectclss): 
            if(elem_type.has_missing):
                seqres = []
                for elem in seq:
                    if(elem is None or elem is Missing):
                        seqres.append(Missing)
                    else: 
                        seqres.append(frozenset(elem))
            else:
                seqres = [frozenset(elem) for elem in seq]
        else:
            seqres = seq
        return util.darray(seqres,elem_type.toNumpy())



class DictConvertor(BaseConvertor):
    def convert(self,seq,elem_type):
        columns = []
        for fieldname, subtype in zip(elem_type.fieldnames, elem_type.subtypes):
            def getname(elem):
                try:
                    return elem[fieldname]
                except (KeyError, TypeError):
                    return Missing
            
            subseq = util.darray([getname(elem) for elem in seq],subtype.toNumpy())
            columns.append(subseq)
        return util.darray(zip(*columns))

class NamedTupleConvertor(BaseConvertor):
    def convert(self,seq,elem_type):
        columns = []
        for fieldname, subtype in zip(elem_type.fieldnames, elem_type.subtypes):
            def getname(elem):
                try:
                    return getattr(elem,fieldname)
                except (KeyError, TypeError):
                    return Missing
            subseq = util.darray([getname(elem) for elem in seq],subtype.toNumpy())
            columns.append(subseq)
        return util.darray(zip(*columns))

class TupleConvertor(BaseConvertor):
    def convert(self,seq,elem_type):
        columns = []
        for pos, subtype in enumerate(elem_type.subtypes):
            itemfunc =  operator.itemgetter(pos)
            def getitem(elem):
                try:
                    return itemfunc(elem)
                except IndexError:
                    return Missing
            subseq = util.darray([getitem(elem) for elem in seq],subtype.toNumpy())
            columns.append(subseq)
        return util.darray(zip(*columns))





class slicetuple(tuple):
    def __new__(self,s):
        res = tuple.__new__(self,(s.start,s.stop,s.step))
        return res

class RPCConvertor(VisitorFactory(prefixes=("execConvert","fconvert"),flags=NF_ELSE)):
    def fconvertobject(self, obj):
        if(hasattr(obj, "__rpc_convert__")):
            return obj.__rpc_convert__()
        else:
            return str(obj)

    def _fnoconvert(self,obj):
        return obj
    fconvertbool = _fnoconvert
    fconvertint = _fnoconvert
    fconvertfloat = _fnoconvert
    fconvertstr = _fnoconvert
    fconvertbuffer = _fnoconvert

    def fconvertinteger(self, obj):
        return int(obj)

    def fconvertfloating(self,obj):
        return float(obj)
    
    def fconvertMissingType(self,obj):
        return None

    def fconvertstring_(self,obj):
        return str(obj)
    
    def fconvertunicode(self,obj):
        return unicode(obj)

    def fconvertbool_(self,obj):
        return bool(obj)

    def fconvertNone(self,obj):
        return None

    def fconverttuple(self, obj):
        return tuple([self.fconvert(elem) for elem in obj])

    def _fconvertsequence(self, obj):
        return [self.fconvert(elem) for elem in obj]
    fconvertlist = _fconvertsequence
    fconvertset = _fconvertsequence

    def fconvertndarray(self, obj):
        if(obj.dtype == object):
            return [self.fconvert(elem) for elem in obj]
        else:
            return obj.tolist()

    def fconvertslice(self, obj):
        return slicetuple(obj)
    
    def fconvertdict(self, obj):
        res = {}
        for key, value in obj.iteritems():
            res[self.fconvert(key)] = self.fconvert(value)
        return res

    def map(self, func, parent_type, seq):
        if(parent_type.has_missing):
            seqres = []
            for elem in seq:
                if(elem is Missing):
                    seqres.append(Missing)
                else: 
                    seqres.append(func(elem))
        else:
            seqres = [func(elem) for elem in seq]
        seqres=util.darray(seqres)
        return seqres 
    

    def execConvertTypeUnknown(self, ptype, seq):
        seq = self.map(self.fconvert, ptype, seq)
        return seq
    
    def execConvertTypeRecordDict(self, ptype, seq):
        if(ptype.has_missing):
            seqres = []
            for elem in seq:
                if elem is Missing:
                    r = Missing
                else:
                    r = dict([(key,self.fconvert(value)) for key,value in elem.iteritems()])
                seqres.append(r)
        else:
            seqres = [dict([(key, self.fconvert(value)) for key,value in elem.iteritems()]) for elem in seq]
        seqres = util.darray(seqres,seq.dtype)
        return seqres
  
    def execConvertTypeSlice(self, ptype, seq):
        return self.map(self.fconvertslice, ptype, seq)

    def execConvertTypeSet(self, ptype, seq):
        return self.map(self.fconvertset, ptype, seq)
   
rpc_convertor = RPCConvertor()   




#from http://code.activestate.com/recipes/267662-table-indentation/

import cStringIO,operator
import numpy
from ..utils import util

def calc_width(rows):
    width = numpy.zeros((len(rows),len(rows[0])),dtype=int)
    for pos,row in enumerate(rows):
        width[pos,:] = map(len,row)
    return width.max(axis=0)
    
def advise_splits(console_width, rows_width, border_width, mincol_width=40):
    nrows_width = rows_width.copy()

    nrows_width[rows_width > mincol_width] = 40

    indices = perform_split(nrows_width, border_width, console_width)
    return indices
    
def optimize_width(console_width, rows_width, border_width):
    nrows_width = rows_width.copy()
    while (nrows_width.sum() + border_width * (len(nrows_width) - 1)) > console_width:
        maxval = nrows_width.max()
        nmax = (nrows_width == maxval).sum()
        rem_lengths = nrows_width[nrows_width != nrows_width.max()]
        tomuch = nrows_width.sum() + border_width * (len(nrows_width) - 1) - console_width
        perrow = numpy.ceil(tomuch / float(nmax))

        if(len(rem_lengths) == 0):
            nrows_width -= perrow
        else:
            next_maxval = rem_lengths.max()
            nrows_width[nrows_width == maxval] -= min(int(perrow), maxval - next_maxval)
    return nrows_width

def perform_split(rows_width, border_width, consolewidth):
    indices = []
    start = 0
    while(start < len(rows_width)):
        stop = start + 1
        while(stop <= len(rows_width) and not (rows_width[start:stop].sum() + border_width * (stop - start - 1)) > consolewidth):
            stop += 1
        indices.append(slice(start,stop))
        start = stop

    

    return indices
    
def row_crop(row_widths, rows):
    nrows = []
    for row in rows:
        nrows.append([crop(col, cwidth) for col, cwidth in zip(row, row_widths)])
    return nrows

def select_cols(rows, indices):
    nrows = []
    for row in rows:
        nrows.append(row[indices])
    return nrows

def prepend_col(rows, col):
    nrows = []
    for elem, row in zip(col, rows):
        nrows.append([elem] + list(row))
    return nrows
    

def indent(rows, hasHeader=False, headerChar='-', delim=' | ', justify='left',
           separateRows=False, prefix='', postfix='', wrapfunc=lambda x:x):
    """Indents a table by column.
       - rows: A sequence of sequences of items, one sequence per row.
       - hasHeader: True if the first row consists of the columns' names.
       - headerChar: Character to be used for the row separator line
         (if hasHeader==True or separateRows==True).
       - delim: The column delimiter.
       - justify: Determines how are data justified in their column. 
         Valid values are 'left','right' and 'center'.
       - separateRows: True if rows are to be separated by a line
         of 'headerChar's.
       - prefix: A string prepended to each printed row.
       - postfix: A string appended to each printed row.
       - wrapfunc: A function f(text) for wrapping text; each element in
         the table is first wrapped by this function."""
    # closure for breaking logical rows to physical, using wrapfunc
    def rowWrapper(row):
        newRows = [wrapfunc(item).split('\n') for item in row]
        if len(row) > 1:
            newRows = map(None, *newRows)
            
        return [[substr or '' for substr in item] for item in newRows]
    # break each logical row into one or more physical ones
    logicalRows = [rowWrapper(row) for row in rows]
    # columns of physical rows
    columns = map(None,*reduce(operator.add,logicalRows))
    # get the maximum of each column by the string length of its items
    maxWidths = [max([len(str(item)) for item in column]) for column in columns]
    rowSeparator = headerChar * (len(prefix) + len(postfix) + sum(maxWidths) + \
                                 len(delim)*(len(maxWidths)-1))
    # select the appropriate justify method
    justify = {'center':str.center, 'right':str.rjust, 'left':str.ljust}[justify.lower()]
    output=cStringIO.StringIO()
    if separateRows: print >> output, rowSeparator
    for physicalRows in logicalRows:
        for row in physicalRows:
            print >> output, \
                prefix \
                + delim.join([justify(str(item),width) for (item,width) in zip(row,maxWidths)]) \
                + postfix
        if separateRows or hasHeader: print >> output, rowSeparator; hasHeader=False
    return output.getvalue()

# written by Mike Brown
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/148061
def wrap_onspace(text, width):
    """
    A word-wrap function that preserves existing line breaks
    and most spaces in the text. Expects that existing line
    breaks are posix newlines (\n).
    """
    return reduce(lambda line, word, width=width: '%s%s%s' %
                  (line,
                   ' \n'[(len(line[line.rfind('\n')+1:])
                         + len(word.split('\n',1)[0]
                              ) >= width)],
                   word),
                  text.split(' ')
                 )

import re
def wrap_onspace_strict(text, width):
    """Similar to wrap_onspace, but enforces the width constraint:
       words longer than width are split."""
    wordRegex = re.compile(r'\S{'+str(width)+r',}')
    return wrap_onspace(wordRegex.sub(lambda m: wrap_always(m.group(),width),text),width)

import math
def wrap_always(text, width):
    """A simple word-wrap function that wraps text on exactly width characters.
       It doesn't split the text in words."""
    return '\n'.join([ text[width*i:width*(i+1)] \
                       for i in xrange(int(math.ceil(1.*len(text)/width))) ])

def crop(text, width):
    if(len(text) > width):
        if(width > 1):
            return text[:(width - 1)] + "~"
        else:
            return text[:width]
    return text

if __name__ == '__main__':
    labels = ('First Name', 'Last Name', 'Age', 'Position')
    data = \
    '''John,Smith,24,Software Engineer
       Mary,Brohowski,23,Sales Manager
       Aristidis,Papageorgopoulos,28,Senior Reseacher'''
    rows = [row.strip().split(',')  for row in data.splitlines()]

    print 'Without wrapping function\n'
    print indent([labels]+rows, hasHeader=True)
    # test indent with different wrapping functions
    width = 10
    for wrapper in (wrap_always,wrap_onspace,wrap_onspace_strict):
        print 'Wrapping function: %s(x,width=%d)\n' % (wrapper.__name__,width)
        print indent([labels]+rows, hasHeader=True, separateRows=True,
                     prefix='| ', postfix=' |',
                     wrapfunc=lambda x: wrapper(x,width))
    
    # output:
    #
    #Without wrapping function
    #
    #First Name | Last Name        | Age | Position         
    #-------------------------------------------------------
    #John       | Smith            | 24  | Software Engineer
    #Mary       | Brohowski        | 23  | Sales Manager    
    #Aristidis  | Papageorgopoulos | 28  | Senior Reseacher 
    #
    #Wrapping function: wrap_always(x,width=10)
    #
    #----------------------------------------------
    #| First Name | Last Name  | Age | Position   |
    #----------------------------------------------
    #| John       | Smith      | 24  | Software E |
    #|            |            |     | ngineer    |
    #----------------------------------------------
    #| Mary       | Brohowski  | 23  | Sales Mana |
    #|            |            |     | ger        |
    #----------------------------------------------
    #| Aristidis  | Papageorgo | 28  | Senior Res |
    #|            | poulos     |     | eacher     |
    #----------------------------------------------
    #
    #Wrapping function: wrap_onspace(x,width=10)
    #
    #---------------------------------------------------
    #| First Name | Last Name        | Age | Position  |
    #---------------------------------------------------
    #| John       | Smith            | 24  | Software  |
    #|            |                  |     | Engineer  |
    #---------------------------------------------------
    #| Mary       | Brohowski        | 23  | Sales     |
    #|            |                  |     | Manager   |
    #---------------------------------------------------
    #| Aristidis  | Papageorgopoulos | 28  | Senior    |
    #|            |                  |     | Reseacher |
    #---------------------------------------------------
    #
    #Wrapping function: wrap_onspace_strict(x,width=10)
    #
    #---------------------------------------------
    #| First Name | Last Name  | Age | Position  |
    #---------------------------------------------
    #| John       | Smith      | 24  | Software  |
    #|            |            |     | Engineer  |
    #---------------------------------------------
    #| Mary       | Brohowski  | 23  | Sales     |
    #|            |            |     | Manager   |
    #---------------------------------------------
    #| Aristidis  | Papageorgo | 28  | Senior    |
    #|            | poulos     |     | Reseacher |
    #---------------------------------------------



import operator

from constants import *
import repops
import repops_funcs

_delay_import_(globals(),"utils","util","context")
_delay_import_(globals(),"ops")
_delay_import_(globals(),"representor")
_delay_import_(globals(),"wrappers","python")
_delay_import_(globals(),"itypes","rtypes","dimensions","dimpaths")
_delay_import_(globals(),"repops_dim")
_delay_import_(globals(),"repops_slice")

class Broadcast(repops.MultiOpRep):
    def _sprocess(self,sources, mode="dim"):
        nslices = []
        for bcslices in util.zip_broadcast(*[source._slices for source in sources]):
            nslices.extend(ops.broadcast(bcslices,mode)[0])
        return self._initialize(tuple(nslices))


class Nest(repops.MultiOpRep):
    def __init__(self, lsource, rsource, dim=LASTCOMMONDIM):
        repops.MultiOpRep.__init__(self,(lsource,rsource),dim=dim)
    
    def _sprocess(self, sources, dim=None):
        assert len(sources) == 2, "Nest expects two representor objects"
        lsource,rsource = sources
        
        joinpath = dimpaths.identifyUniqueDimPathSource(lsource, dim)
        joinpath = dimpaths.extendParentDim(joinpath,[s.dims for s in source._slices], ALL)

        idims = []
        for i in xrange(len(joinpath)):
            idims.append(dimensions.Dim(1))
        
        
        references = []
        for ndim in joinpath:
            nrefs = []
            for slice in lsource._slices:
                if ndim in slice.dims:
                    nrefs.append(slice)
            references.append(nrefs)

        nslices = []
        plan = [BCEXIST] * len(idims)
        for slice in rsource._slices:
            odims = slice.dims
            for dimpos in xrange(len(joinpath)):
                slice = ops.InsertDimOp(slice,dimpos,idims[dimpos])
            slice = ops.BroadcastOp(slice,references,plan,joinpath + odims)
            nslices.append(slice)
        return self._initialize(tuple(nslices))

class Combine(repops.MultiOpRep):
    def __init__(self, *sources):
        repops.MultiOpRep.__init__(self, sources)
    def _sprocess(self,sources):
        nslices = self._apply(*[source._slices for source in sources])
        return self._initialize(nslices)

    @classmethod
    def _apply(cls, *xslicelists):
        if(len(xslicelists) == 2):
            lslices,rslices = xslicelists
            lslices = [ops.ChangeBookmarkOp(lslice,add_bookmark="!L",update_auto_bookmarks="L") for lslice in lslices]
            rslices = [ops.ChangeBookmarkOp(rslice,add_bookmark="!R",update_auto_bookmarks="R") for rslice in rslices]
            return tuple(lslices + rslices)
        else:
            return sum([tuple(xslices) for xslices in xslicelists],tuple())

class Group(repops.MultiOpRep):
    def __init__(self,source,constraint,flat={}):
        repops.MultiOpRep.__init__(self,(source,constraint),flat=flat)

    def _process(self,sources, flat):
        source, gsource = sources 
        if not source._slicesKnown() or not gsource._typesKnown():
            return

        gslices = [ops.ensure_frozen(slice) for slice in gsource._slices]
        gslices = ops.broadcast(gslices,mode="dim")[0]
        gslices = [ops.PackArrayOp(gslice,1) for gslice in gslices]

        gslice = ops.GroupIndexOp(gslices)
        gslice = ops.UnpackArrayOp(gslice, len(gslices))

        #determine which slices to flatten
        oslices = list(source._slices)
        nslices = [None] * len(oslices)
        xflat = {}
        #specified slices
        for key,values in flat.iteritems():
            if(isinstance(key,tuple)):
                keyslices = gsource.Get(*key)
            else:
                keyslices = gsource.Get(key)
            nkey = tuple([gsource._slices.index(keyslice) for keyslice in keyslices._slices])
             
            if(isinstance(values,(list,tuple))):
                vslices = source.Get(*values)
            else:
                vslices = source.Get(values)
            
            pos = []
            for slice in vslices._slices:
                assert slice in oslices, "Cannot find slice specified in group flat argument"
                p = oslices.index(slice)
                pos.append(p)
                oslices[p] = None
            xflat[nkey] = set(pos)

        #group slices
        for pos, slice in enumerate(gsource._slices):
            try:
                p = oslices.index(slice)
            except ValueError:
                continue
            key = (pos,)
            if(key in xflat):
                xflat[key].add(p)
            else:
                xflat[key] = set([p])
            oslices[p] = None

        #filter remaining slices first
        pos = [p for p,s in enumerate(oslices) if not s is None]
        if(pos):
            fslices = [s for s in oslices if not s is None]
            fslices = Filter._apply(fslices,gslice,gslices[0].type.dims[:1],"dim")
            for p,fslice in zip(pos,fslices):
                nslices[p] = fslice

        #filter flattened slices 
        gidx = range(len(gsource._slices))
        gdims = gslice.dims[-len(gsource._slices):]
        firstelem = python.Rep(0,name="data")._slices[0]
        for key, pos in xflat.iteritems():
            tslice = gslice
            for elem in gidx:
                if not elem in key:
                    selpath = dimpaths.DimPath(gdims[elem])
                    tslice = repops_funcs.Sum._apply([tslice],selpath)[0]
            tslice = ops.UnpackArrayOp(tslice,1)
            tslice = Filter._apply([tslice], firstelem,tslice.dims,"dim")[0]
            tslice = ops.PackArrayOp(tslice,1)
            pos = list(pos)
            fslices = [source._slices[p] for p in pos]
            fslices = Filter._apply(fslices,tslice,gslices[0].type.dims[:1],"dim")
            for p,fslice in zip(pos,fslices):
                nslices[p] = fslice
            
        return self._initialize(tuple(nslices))


class Filter(repops.MultiOpRep):
    def __init__(self,source,constraint,dim=LASTCOMMONDIM,mode=None):
        if(not isinstance(constraint,representor.Representor)):
            constraint = repops.PlusPrefix(python.Rep(constraint,name="filter"))
        repops.MultiOpRep.__init__(self,(source,constraint),dim=dim,mode=mode)

    def _process(self,sources,dim,mode):
        source,constraint = sources
        if not source._slicesKnown() or not constraint._typesKnown():
            return
        assert len(constraint._slices) == 1, "Filter constraint should have 1 slice"
        cslice = constraint._slices[0]

        if(not isinstance(cslice.type, rtypes.TypeBool)):
            seldimpath = dimpaths.identifyUniqueDimPathSource(source, dim)
        else:
            seldimpath = dimpaths.identifyUniqueDimPathSource(source, cslice.dims[-1])
            
        if(not seldimpath and not isinstance(cslice.type, rtypes.TypeBool)):
            raise RuntimeError, "Attempting to perform filter on non-existing dimension"
        
        if mode is None:
            if(isinstance(constraint, repops.PlusPrefix)):
                mode = "pos"
            else:
                mode = "dim"
           
        nslices = self._apply(source._slices,cslice,seldimpath,mode)
        return self._initialize(tuple(nslices))

    @classmethod
    def _apply(cls,fslices,cslice,seldimpath, bcmode):
        if(isinstance(cslice.type,rtypes.TypeBool)):
            assert cslice.dims, "Constraint should have at least one dimension"
            ndim = dimensions.Dim(UNDEFINED,(True,) * (len(cslice.dims) -1),  False, name = "f" + cslice.dims[-1].name)
            dim_suffix = None
            seldimpath = cslice.dims[-1:]
            dimset = dimpaths.createDimSet([slice.dims for slice in fslices])
            assert seldimpath[0] in dimset, "Cannot find last dimension of boolean filter in filter source (" + str(cslice.dims) + ")"
            cslice = ops.PackArrayOp(cslice,1)
        else:
            assert seldimpath, "Filter dimpath is empty"

            if(isinstance(cslice.type, rtypes.TypeInteger)):
                ndim = None
            elif(isinstance(cslice.type, rtypes.TypeArray)):
                assert len(cslice.type.dims) == 1, "Filter array should be 1-dimensional"
                assert isinstance(cslice.type.subtypes[0], rtypes.TypeInteger) and \
                            not isinstance(cslice.type.subtypes[0], rtypes.TypeBool), \
                            "Multi-dimensional arrays cannot be used as filter. Please unpack the arrays."
                ndim = cslice.type.dims[0]
            elif(isinstance(cslice.type, rtypes.TypeSlice)):
                ndim = dimensions.Dim(UNDEFINED, (True,) * len(cslice.dims), False,name = "f" + seldimpath.strip()[-1].name)
            else:
                raise RuntimeError, "Unknown constraint type in filter: " + str(cslice.type)
        
        nslices = []
        for slice in fslices:
            slice = ops.filter(slice, cslice, seldimpath, ndim, bcmode)
            nslices.append(slice)
        return nslices

class Sort(repops.MultiOpRep):
    def __init__(self,source,constraint=None,descend=False):
        if(constraint is None):
            repops.MultiOpRep.__init__(self,(source,),descend=descend)
        else:
            if(not isinstance(constraint,representor.Representor)):
                constraint = repops.PlusPrefix(python.Rep(constraint,name="filter"))
            repops.MultiOpRep.__init__(self,(source,constraint),descend=descend)

    def _sprocess(self, sources, descend):
        if(len(sources) == 1): #no explicit constraint, use data itself
            source = sources[0]
            if len(source._slices) > 1:
                constraint = source.Tuple()
            else:
                constraint = source
        else:
            source,constraint = sources 

        if len(constraint._slices) > 1:
            constraint = constraint.Tuple()
        
        #fixme: make it slice-only (remove rep)
        constraint = repops_funcs.Argsort(constraint, descend=descend)
        cslice = ops.PackArrayOp(constraint._slices[0])
        nslices = Filter._apply(source._slices, cslice, cslice.type.dims[:1],"dim")
        return self._initialize(tuple(nslices))


class Unique(repops.MultiOpRep):
    def __init__(self,source,constraint=None,descend=False):
        if(constraint is None):
            repops.MultiOpRep.__init__(self,(source,),descend=descend)
        else:
            if(not isinstance(constraint,representor.Representor)):
                constraint = repops.PlusPrefix(python.Rep(constraint,name="filter"))
            repops.MultiOpRep.__init__(self,(source,constraint),descend=descend)

    def _process(self, sources, descend):
        if(len(sources) == 1): #no explicit constraint, use data itself
            source = sources[0]
            if len(source._slices) > 1:
                constraint = source.Tuple()
            else:
                constraint = source
        else:
            source,constraint = sources 
        if not source._slicesKnown() or not constraint._typesKnown():
            return

        if len(constraint._slices) > 1:
            constraint = constraint.Tuple()
        
        #fixme: make it slice-only (remove rep)
        dpath = dimpaths.identifyUniqueDimPathSource(constraint,LASTCOMMONDIM)
        uconstraint = repops_funcs.Argunique(constraint)
        cslice = uconstraint._slices[0]
        nslices = Filter._apply(source._slices, cslice, dpath[-1:],"dim")
        return self._initialize(tuple(nslices))


class Join(repops.MultiOpRep):
    def __init__(self, lsource, rsource, constraint):
        if(isinstance(constraint,context.Context)):
            c = Combine(lsource,rsource)
            constraint = context._apply(constraint,c)
        repops.MultiOpRep.__init__(self,(lsource,rsource, constraint))

    def _process(self, sources):
        lsource, rsource, constraint = sources
        if not lsource._slicesKnown() or not rsource._slicesKnown() or not constraint._typesKnown():
            return
        
        ldims = dimpaths.createDimSet([s.dims for s in lsource._slices])
        rdims = dimpaths.createDimSet([s.dims for s in rsource._slices])
        
        assert len(constraint._slices) == 1, "Constraint should have only one slice"
        cslice = constraint._slices[0]

        if not cslice.dims[-1] in rdims:
            rdims,ldims = ldims,rdims
            lsource,rsource = rsource,lsource
            reverse = True
            assert cslice.dims[-1] in rdims, "Last dimension of constraint should be in one of the sources"
        else:
            reverse = False
        rdimpath = dimpaths.DimPath(cslice.dims[-1])
        for dim in cslice.dims[::-1]:
            if dim in ldims:
                ldimpath = dimpaths.DimPath(dim)
                break
        else:
            raise RuntimeError, "No dimension found for one of the sources in constraint"

        leftpos = repops_funcs.Pos._apply([cslice], ldimpath)[0]
        rightpos = repops_funcs.Pos._apply([cslice], rdimpath)[0]
        
        filters = Filter._apply([leftpos,rightpos],cslice, LASTCOMMONDIM,"dim")
        leftflat,rightflat = repops_dim.Flat._apply(filters, filters[0].dims[-1:])

        leftflat = ops.PackArrayOp(leftflat)
        rightflat = ops.PackArrayOp(rightflat)

        lslices = Filter._apply(lsource._slices, leftflat, ldimpath, "dim")
        rslices = Filter._apply(rsource._slices, rightflat, rdimpath, "dim")

        if(reverse):
            nslices = Combine._apply(rslices,lslices)
        else:
            nslices = Combine._apply(lslices,rslices)
        return self._initialize(tuple(nslices))



class Match(repops.MultiOpRep):
    def __init__(self, lsource, rsource, lslice=None, rslice=None, jointype="inner", merge_same=False, mode="dim"):
        assert jointype in set(["inner","left","right","full"]), "Jointype should be inner, left, right or full"
        assert (not isinstance(lslice,representor.Representor)), "Representor objects not allowed as lslice. Use context, string or int to indicate slice in lsource"
        assert (not isinstance(rslice,representor.Representor)), "Representor objects not allowed as rslice. Use context, string or int to indicate slice in rsource"
        
        if not rslice is None:
            rslice = rsource.Get(rslice)
        elif not lslice is None:
            rslice = rsource.Get(lslice)

        if not lslice is None:
            lslice = lsource.Get(lslice)
            
        repops.MultiOpRep.__init__(self,(lsource,rsource,lslice,rslice),jointype=jointype, merge_same=merge_same,mode=mode)

    def _process(self, sources, jointype, merge_same,mode):
        lsource, rsource, lslice,rslice = sources
        if not lsource._slicesKnown() or not rsource._slicesKnown():
            return
        if(lslice is None):
            if(rslice is None):
                common_names = set(lsource.Names) & set(rsource.Names)
                if(len(common_names) != 1):
                    raise RuntimeError, "Cannot find a unique common named slice"
                name = common_names.pop()
                lslice = getattr(lsource,name)
                rslice = getattr(rsource,name)
            else:
                lslice = getattr(lsource,rslice.Names[0])
        elif(rslice is None):
            rslice = getattr(rsource,lslice.Names[0])
        if not lslice._slicesKnown() or not rslice._slicesKnown():
            return
        assert len(rslice._slices) == 1, "rslice parameter in match should have only one slice"
        assert len(lslice._slices) == 1, "lslice parameter in match should have only one slice"
        
        assert merge_same in set([False, 'equi','all',True]) or isinstance(merge_same,tuple), 'merge_same should be "equi", "all"/True or a tuple of names, or a tuple of tuple of name pairs'

        self._sources = (lsource, rsource, lslice, rslice)
        if not lslice._typesKnown() or not rslice._typesKnown():
            return
        lslice = lslice._slices[0]
        rslice = rslice._slices[0]
        nslices = self._apply(lsource, rsource, lslice, rslice, jointype, merge_same, mode)
        return self._initialize(tuple(nslices))


    @classmethod
    def _apply(cls,lsource, rsource, lslice,rslice, jointype="inner", merge_same=False, mode="dim"):
        lslice = ops.ensure_frozen(lslice)
        rslice = ops.ensure_frozen(rslice)

        leftslice = ops.PackArrayOp(lslice)
        rightslice = ops.PackArrayOp(rslice)
       
        ((bleftslice,brightslice),(leftplan,rightplan))= ops.broadcast((leftslice,rightslice),mode=mode)

        lindex,rindex = ops.EquiJoinIndexOp(bleftslice,brightslice, jointype=jointype, mode=mode).results

        lslices = list(lsource._slices)
        rslices = list(rsource._slices)
        
        if isinstance(merge_same, tuple):
            nms = dict()
            for ms in merge_same:
                if not isinstance(ms,tuple):
                    nms[ms] = ms
                else:
                    nms[ms[0]] = ms[1]
            merge_same = nms

        collapse_equi = (merge_same == 'equi' or merge_same == 'all' or merge_same is True) or (isinstance(merge_same, dict) and lslice.name in merge_same and rslice.name in merge_same[lslice.name])

        if((lslice.name == rslice.name or collapse_equi) and lslice in lslices and rslice in rslices):
            if jointype== 'inner':                
                rslices.pop(rslices.index(rslice))
            if collapse_equi:
                if jointype == 'left':
                    rslices.pop(rslices.index(rslice))
                elif jointype == 'right':
                    lslices.pop(lslices.index(lslice))
                else:
                    lidx = lslices.index(lslice)
                    ridx = rslices.index(rslice)

        lslices = [ops.broadcastParentsFromPlan(tslice, lslice.dims[-1:], leftplan, leftslice.dims, bleftslice.dims, [rightslice], True) for tslice in lslices]
        rslices = [ops.broadcastParentsFromPlan(tslice, rslice.dims[-1:], rightplan, rightslice.dims, brightslice.dims, [leftslice], True) for tslice in rslices]

        lslices = list(Filter._apply(lslices, lindex, lslice.dims[-1:], "dim"))
        rslices = list(Filter._apply(rslices, rindex, rslice.dims[-1:], "dim"))

        if collapse_equi and jointype == 'full':
            lslice = lslices[lidx]
            rslice = rslices[ridx]
            lslices[lidx] = repops_funcs.Merge._apply([lslice,rslice], "dim")
            del rslices[ridx]

        
        if merge_same == 'all' or merge_same is True:
            ms = dict()
            lnames = [lslice.name for lslice in lslices]
            for rslice in list(rslices):
                if not rslice.name in lnames:
                    continue
                ms[rslice.name] = rslice.name                    
            merge_same = ms

        if isinstance(merge_same, dict):
            lnames = [lslice.name for lslice in lslices]
            rnames = [rslice.name for rslice in rslices]
            for lname, rname in merge_same.iteritems():
                if lname in lnames and rname in rnames:
                    lpos = lnames.index(lname)
                    rpos = rnames.index(rname)
                    lslices[lpos] = repops_funcs.Merge._apply([lslices[lpos],rslices[rpos]], "dim")
                    del rslices[rpos]

        return Combine._apply(lslices,rslices)

class Replace(repops.MultiOpRep):
    def __init__(self, source, slice, translator, fromslice=0, toslice=1):
        slice = source.Get(slice)
        fromslice = translator.Get(fromslice)
        toslice = translator.Get(toslice)
        translator = Combine(fromslice, toslice)

        repops.MultiOpRep.__init__(self,(source, slice, translator))
   
    def _sprocess(self, sources):
        source, slice, translator = sources
        assert len(slice._slices) == 1, "Only one slice can be replaced"
        assert len(translator._slices) == 2, "Translator should have only two selected slices"

        slice = slice._slices[0]
        slicepos = source._slices.index(slice)

        nslices = list(Match._apply(source, translator, slice, translator._slices[0],"inner"))
        nslices[slicepos] = ops.ChangeNameOp(nslices[-1],slice.name)
        remove = (len(source._slices) + len(translator._slices)) -len(nslices) - len(translator._slices)
        nslices = nslices[:remove]
        return self._initialize(tuple(nslices))


class Take(repops.MultiOpRep):
    def __init__(self, source, take_source, allow_missing=False, keep_missing=False):
        if (isinstance(source,dict)):
             source = python.Rep(list(source.iteritems()),name="data").IndexDict()
        repops.MultiOpRep.__init__(self,(source, take_source), allow_missing = allow_missing, keep_missing=keep_missing)
   
    def _sprocess(self, sources, allow_missing,keep_missing):
        source, take_source = sources
        if len(source._slices) == 2:
            source = source.IndexDict()
        assert len(source._slices) == 1, "Take source should have one slice"
        source_slice = source._slices[0]

        nslices = []
        for take_slice in take_source._slices:
            nslices.append(ops.TakeOp(source_slice, take_slice, allow_missing, keep_missing))
        return self._initialize(tuple(nslices))


class Stack(repops.MultiOpRep):
    def __init__(self, *sources, **kwargs):
        repops.MultiOpRep.__init__(self,sources, **kwargs)

    def _sprocess(self, sources, slices=COMMON_POS, mode='dim', dims=LCDIM):
        nslices = []
        if slices == COMMON_POS:
            slicelens = set([len(source._slices) for source in sources])
            assert len(slicelens) == 1, ("Sources should have same number of slices in %s when matching based on COMMON_POS" % self.__class__.__name__ )
        elif slices == COMMON_NAME:
            nms = set(sources[0].Names)
            assert len(sources[0].Names) == len(nms), ('Cannot have duplicate names in %s while matching slices on name' % self.__class__.__name__)
            for source in sources[1:]:
                xnms =  set(source.Names)
                assert len(source.Names) == len(xnms), ('Cannot have duplicate names in %s while matching slices on name' % self.__class__.__name__)
                nms &= xnms
            assert nms, ('Cannot find shared names in %s while matching slices on name' % self.__class__.__name__)

            nsources = []
            for source in sources:
                onms = set(source.Names) - nms
                if onms:
                    nslices.extend(source.Get(*onms)._slices)
                nsources.append(source.Get(*nms))
            sources = nsources
        
        seldimpaths = [] 
        if not isinstance(dims,tuple):
            dims = (dims,) * len(sources)
        for source, dim  in zip(sources, dims):
            seldimpaths.append(dimpaths.identifyUniqueDimPathSource(source, dim))

        slicelists = [source._slices for source in sources]
        ndim = None
        for slicecol in zip(*slicelists):
            packdepths = []
            ncol = []
            for slice,dpath in zip(slicecol, seldimpaths):
                lastpos = slice.dims.matchDimPath(dpath)
                assert len(lastpos) == 1, "Cannot choose between or find dims in slice: " + str(slice)
                packdepths.append(len(slice.dims) - lastpos[0])
                if(packdepths[-1] > 0):
                    slice = ops.PackArrayOp(slice, packdepths[-1])
                ncol.append(slice)
            
            #assert len(set([len(s.dims) for s in ncol])) == 1, "Dim depth mismatch between slices"

            res = ncol[0]
            for slice in ncol[1:]:
                res = repops_funcs.Add._apply((res,slice),mode)
            dimdepth = min(packdepths)
            res = ops.UnpackArrayOp(res,dimdepth)
            
            if ndim is None:
                ndim = res.dims[-dimdepth]
            else:
                res = ops.ChangeDimOp(res, len(res.dims) - dimdepth, ndim)
            nslices.append(res)

        return self._initialize(tuple(nslices))


class Intersect(repops.MultiOpRep):
    opercls = repops_funcs.And

    def __init__(self, *sources, **kwargs):
        repops.MultiOpRep.__init__(self,sources, **kwargs)

    def _sprocess(self, sources, slices=COMMON_POS, dims=LCDIM, mode='dim'):
        if any([not source._typesKnown() for source in sources]):
            return

        nslices = []
        if slices == COMMON_POS:
            slicelens = set([len(source._slices) for source in sources])
            assert len(slicelens) == 1, ("Sources should have same number of slices in %s when matching based on COMMON_POS" % self.__class__.__name__ )
        elif slices == COMMON_NAME:
            nms = set(sources[0].Names)
            assert len(sources[0].Names) == len(nms), ('Cannot have duplicate names in %s while matching slices on name' % self.__class__.__name__)
            for source in sources[1:]:
                xnms =  set(source.Names)
                assert len(source.Names) == len(xnms), ('Cannot have duplicate names in %s while matching slices on name' % self.__class__.__name__)
                nms &= xnms
            assert nms, ('Cannot find shared names in %s while matching slices on name' % self.__class__.__name__)

            nsources = []
            for source in sources:
                onms = set(source.Names) - nms
                if onms:
                    nslices.extend(source.Get(*onms)._slices)
                nsources.append(source.Get(*nms))
            sources = nsources
        elif isinstance(slices, tuple):
            assert len(slices) == len(sources), ('Slices tuple argument to %s should have same number of element as their are source datasets' % self.__class__.__name__)
            
            nsources = []
            for source, xslice in zip(sources, slices):
                if not isinstance(xslice, tuple):
                    xslice = (xslice,)
                selsource = source.Get(*xslice)                    
                noselsource = source.Without(selsource)
                nslices.extend(noselsource._slices)
                nsources.append(selsource)
            sources = nsources                
            
        seldimpaths = [] 
        tupleslices = []
        packdepthslist = []
        if not isinstance(dims,tuple):
            dims = (dims,) * len(sources)

        for dim, source in zip(dims, sources):
            dpath = dimpaths.identifyUniqueDimPathSource(source, dim)
            pslices = []
            packdepths = []
            for slice in source._slices:
                lastpos = slice.dims.matchDimPath(dpath)
                if not lastpos:
                    nslices.append(slice)
                else:
                    packdepths.append(len(slice.dims) - lastpos[-1] - 1)
                    assert len(lastpos) == 1, "Cannot find dims in slice: " + str(slice)
                    if(packdepths[-1] > 0):
                        slice = ops.PackArrayOp(slice, packdepths[-1])
                    pslices.append(slice)
            assert len(set([len(s.dims) for s in pslices])) == 1, "Dim depth mismatch between slices in a source"
            if(len(pslices) > 1):
                pslice = repops_slice.Tuple._apply(pslices)
            else:
                pslice = pslices[0]
            pslice = repops_funcs.Set._apply([pslice], None)[0]
            tupleslices.append(pslice)
            packdepthslist.append(packdepths)
        
        res = tupleslices[0]
        for slice in tupleslices[1:]:
            res = self.opercls._apply((res,slice),mode)

        tslice = ops.UnpackArrayOp(res,1)
        if(isinstance(tslice.type, rtypes.TypeTuple)):
            pslices = [ops.UnpackTupleOp(tslice, idx) for idx in range(len(tslice.type.subtypes))]
        else:
            pslices = [tslice]
       
        for pos, pslice in enumerate(pslices):
            mpd = min([pd[pos] for pd in packdepthslist])
            if(mpd > 0):
                pslice = ops.UnpackArrayOp(pslice, mpd)
            nslices.append(pslice)

        return self._initialize(tuple(nslices))

class Union(Intersect):
    opercls = repops_funcs.Or


class Except(Intersect):
    opercls = repops_funcs.Subtract


class Difference(Intersect):
    opercls = repops_funcs.Xor




import copy
from constants import *
import representor
_delay_import_(globals(),"ops")
_delay_import_(globals(),"itypes","rtypes")
_delay_import_(globals(),"itypes","dimpaths")
_delay_import_(globals(),"utils","context")
_delay_import_(globals(),"repops_multi")

def delayable(default_params=["*"], nsources=1):
    """Function to enable the delay of single source operations.

    default_params: slice to apply the function to if no parameter is
                   given.

    Delayable functions should have as first param a source field. 
    If source is:
    - a Representor object: functions like normal function
    - a Context object: creates a context object with the function applied
    - any other: creates a context object with the parameter for source 
      given to the get function of the representor the context object
      should be applied to.
    - none: replaces with the argument in default_slice
    Other parameters of the function can still be given, either by 
    param order or keyword.
    Examples:
    data = Rep(data) (f1, f2, f3 slices)

    data[pos(data.f0) < 10] == data[pos(_.f0) < 10] == \
        data[pos("#") < 10] == data[pos() < 10]

    """
    def new(func):
        def delayable_function(*params, **kwds):
            if(not params):
                return _.Get(*default_params)._call(func, **kwds)
            
            if(nsources == UNDEFINED):
                xnsources = len(params)
            else:
                xnsources = nsources

            need_get = False

            for i in xrange(xnsources):
                if(not isinstance(params[i], representor.Representor)):
                    need_get = True
            if(need_get):
                return _.Get(*params[:xnsources])._call(func, *params[xnsources:], **kwds)
            else:
                if(xnsources > 1):
                    return func(repops_multi.Combine(*params[:xnsources]), *params[xnsources:], **kwds)
                else:
                    return func(*params, **kwds)
        return delayable_function
    return new

class UnaryOpRep(representor.Representor):
    def __init__(self, source, *args, **kwds):
        assert isinstance(source,representor.Representor), "Source should be a representor"
        self._source = source
        self._params = (args,kwds)
        self._process(source,*args, **kwds)

    def _process(self, source, *args, **kwds):
        if not source._slicesKnown():
            return
        return self._sprocess(source, *args, **kwds)

    def _sprocess(self, source, *args, **kwds):
        return self._initialize(source._slices)

class MultiOpRep(representor.Representor):
    def __init__(self, sources, *args,**kwds):
        assert isinstance(sources,tuple), "Sources should be a tuple"
        self._sources = sources
        self._params = (args,kwds)
        self._process(sources,*args, **kwds)
    
    def _process(self, sources, *args, **kwds):
        if not all([source._slicesKnown() for source in sources]):
            return
        return self._sprocess(sources, *args, **kwds)      
    
    def _sprocess(self, sources,*args, **kwds):
        raise RuntimeError, "Process function should be overloaded for " + str(type(self))
    

class Fixate(UnaryOpRep):#{{{
    """Operation used by optimizer to fixate end of tree,
    such that there are no exception situations, and slice retrieval
    is handled correctly."""

    def _sprocess(self, source):
        nslice = ops.FixateOp(source._slices)
        self._initialize((nslice,))
    #}}}

class Gather(Fixate):#{{{
    """Operation used by optimizer to fixate end of tree,
    such that there are no exception situations, and slice retrieval
    is handled correctly."""
    def _sprocess(self, source):
        nslice = ops.GatherOp(source._slices)
        self._initialize((nslice,))#}}}

class PlusPrefix(UnaryOpRep):#{{{
    pass
#}}}

class NoOp(UnaryOpRep):#{{{
    pass
#}}}

class ApplyFuncRep(UnaryOpRep):
    """Applies slice class in `slicecls` to every field in source.

    :param dim_selector: select dimensions on which to apply the slicecls
    """

    def _sprocess(self,source,func,*params,**kwds):
        nslices = func(source._slices, *params, **kwds)
        return self._initialize(nslices)


class Detect(UnaryOpRep):
    """Detect types
    """

    def _sprocess(self,source, *args, **kwargs):
        only_unknown = kwargs.pop('only_unknown',False)
        allow_convert = kwargs.pop('allow_convert',True)
        if only_unknown:
            nslices = []
            for slice in source._slices:
                if slice.type == rtypes.unknown:
                    nslices.append(ops.DetectAndCastOp(slice, allow_convert=allow_convert))
                else:
                    nslices.append(slice)
        else:
            nslices = [ops.DetectAndCastOp(slice, allow_convert=allow_convert) for slice in source._slices]
        return self._initialize(tuple(nslices))


def apply_slice(slices, slicecls, dim_selector, *params, **kwds):
    """Applies slice class in `slicecls` to every field in source.

    :param dim_selector: select dimensions on which to apply the slicecls
    """
    if(not dim_selector is None):
        dim_selector = dimpaths.identifyUniqueDimPath(slices, dim_selector)
        nslices = []                       #new active slices
        for slice in slices:
            if(dim_selector in slice.dims):
                nslice = slicecls(slice, *params, **kwds)
            else:
                nslice = slice
            nslices.append(nslice)
    else:
        nslices = [slicecls(slice,*params,**kwds) for slice in slices]
    return tuple(nslices)





"""Detector and scanner code. Used to determine data types.

When importing data, types can be represented in multiple ways. 
An integer can be represented as a string, etc. These have 
to be detected (and converted to a standard format). In this module, 
there is a general Detector class, which handles Scanner classes, which
do the actual type checking. 
Main workflow:

d = Detector()
d.process(<your data obj>)
[d.process(<your data obj>)]
d.getType()

.. rubric:: Helper classes
    TypeScanner (derivatives): Scan for specific types.
    DimRep:                    Process and store dimension shapes
    DimensionEqualizer:        Find common dimensions


.. rubric:: Overview
    Scanners follow a hierarchy, starting with the AnyScanner, which has
    children such as CollectionScanner, TupleScanner and NumberScanner. 
    Each child can have children of his own. 

    Each detector first performs a depth-first search of allowable scanners,
    and then during data processing goes up in the hiearchy if the more specific
    children-scanners cannot match the data anymore.

    Scanners implement a 'scan' method, which gets a sequence of objects, and
    should return False/True if these are objects are acceptable. Detector performs
    a common detection step in that it determines a set of class types available in the
    sequence. 

    Once the user asks for the type, the detector asks every active scanner for a type, and
    next will (try) to determine which type should be returned. For this, the scanner class
    should have a getType method, where it creates an type based on the collected information.

    During the data scanning, scanners will sometimes not be able to match the data, after
    which their parent scanner is activated. Earlier data is then not scanned again, instead
    the child scanner should transfer its collected information to the parent scanner during
    its own unregistration (the 'unregister' method will be called with create_parent flag set to 
    True). 

    If a scanner has to scan subtypes, it should open a subdetector class (for which one 
    can use the getSubDetector method implemented in typescanner). This detector should than be 
    feeded the subobjects. 

    To determine dimensions, one can use the DimensionEqualizer/DimRep classes. Each dimension 
    should have its own DimRep class, which can process dimensions lengths using the processLengths
    method. It will store if this length is fixed/variable/repetitive, and allow for comparison between
    dimensions. One can open a new DimRep class with getDimRep method. If during scanning it becomes clear
    that the number of dimensions has to be reduced, one can do this with the reduceDimReps method. 

    During the call to getType, DimensionEqualizer will compare DimReps, and assign dimension objects
    to each. These can be obtained from the .dim attributed during type creation. 


.. rubric:: Adding a new scanner class
    The scanner hierarchy is separate from the class hierarchy. So one should find the class which has 
    the most common implementation (often just TypeScanner).
    Next, one should implement the scan method. See for examples below.
    The parentcls attribute should be used to determine the parent scanner,
    the registerTypeScanner method to register the scanner. 
"""

import numpy
import array
from collections import defaultdict
import operator
import platform
from itertools import chain

import rtypes
from ..constants import *
from ..utils import sparse_arrays, module_types
from ..utils.missing import *

_delay_import_(globals(),"dimensions")
_delay_import_(globals(),"dimpaths")
_delay_import_(globals(),"..utils","cutils","util")

_scanchildren = defaultdict(list)
def registerTypeScanner(newscancls):
    _scanchildren[newscancls.parentcls].append(newscancls)

missing_cls = set([MissingType])
missing_val = set([Missing])


class Detector(object):
    def __init__(self, parent_scanner=None, dim_eq=None, allow_need_convert=False):
        self._scanners = None
        self.objectclss = set()
        self.count_elem = 0
        self.allow_need_convert = allow_need_convert

        if not dim_eq:
            dim_eq = DimEqualizer()

        self.dim_eq = dim_eq
        self.parent_scanner = parent_scanner
        return None

    def setParentDimensions(self,dims):
        assert self.count_elem == 0, "Cannot perform setParentDimensions after process(Seq) has been called"
        self._scanners = [OuterContainerScanner(self,dims)]

    def process(self, obj):
        self.processSeq(sparse_arrays.FullSparse(util.darray([obj])))

    def processSeq(self, seq):
        if not isinstance(seq, (sparse_arrays.FullSparse)):
            if isinstance(seq, (set, frozenset)):
                seq = util.darray(list(seq))
            elif not isinstance(seq, numpy.ndarray):
                seq = util.darray(seq)
            seq = sparse_arrays.FullSparse(seq)
        
        objclasses = seq.classes
        self.objectclss |= objclasses
        elem_count = len(seq)

        if not objclasses.issubset(missing_cls):
            if not self._scanners:
                test_seq = seq[:100]
                seq = seq[100:]
                self._scanners = list(self._findAcceptableDescendants(test_seq))
            if len(seq) > 0:
                for pos in xrange(len(self._scanners) - 1, -1, -1):
                    scanner = self._scanners[pos]
                    while not scanner.scan(seq):
                        del self._scanners[pos]
                        if not any([scanner.parentcls in s.ancestor_scanners for s in self._scanners]):
                            scanner = scanner.unregister(create_parent=True)
                            self._scanners.insert(pos,scanner)
                            continue
                        scanner.unregister()
                        break
        self.count_elem += elem_count

    def getType(self):
        if not self._scanners:
            return rtypes.TypeUnknown()
        if not self.parent_scanner:
            self.dim_eq.processDims()

        restypes = [scanner.getType() for scanner in self._scanners]
        
        restypes = [t for t in restypes if t is not None]
        
        assert not len(restypes) == 0,'BUG: no valid type could be detected'

        if len(restypes) > 1:
            raise RuntimeError, 'Could not decide between types: ' + str(restypes)
        return restypes[0]

    def _findAcceptableDescendants(self, seq, scanner=None):
        schildren = [sc(self) for sc in _scanchildren[scanner.__class__] if (sc.need_convert == True and self.allow_need_convert == True) or sc.need_convert == False]
       
        res = set()
        for sc in schildren:
            if sc.scan(seq):
                res.update(self._findAcceptableDescendants(seq, sc))
            else:
                sc.unregister()
        
        if not res:
            return (scanner,)

        if scanner:
            scanner.unregister()

        return res

    def hasMissing(self):
        return not missing_cls.isdisjoint(self.objectclss)



LENGTH_NOINIT = 0    #length type not yet initialized
LENGTH_FIXED = 1     #lengths are fixed (all the same or missing)
LENGTH_VAR = 2       #variable lengths
LENGTH_REPEAT = 3    #lengths repeated in a pattern dictated by parent dimensions


class DimRep(object):
    def __init__(self, parent_scanner, index=0):
        self.parent_scanner = parent_scanner
        self.index = index

        self.lengths = None        #stores length data. FIXED: int, VAR: list of lengths, REPEAT: list of lengths
        self.length_count = 0      #FIXED: stores number of elements, REPEAT: number of repeats
        self.length_type = LENGTH_NOINIT

        self.nparents = 0

        self.has_missing = False
        self.dim = None
        self.dirty = True

    def processLengths(self, lengths, has_missing=False):
        self.has_missing = max(self.has_missing, has_missing)
        self.dirty = True

        if has_missing:
            lenghts = lengths.replace_missing(-1, otype=int)

        if self.length_type == LENGTH_FIXED or self.length_type == LENGTH_NOINIT:
            lenset = set(lengths)
            lenset.discard(-1)
            if len(lenset) == 0:
                self.length_count += len(lengths)
            elif len(lenset) > 1:
                self._fullLengths()
                self.processLengths(lengths)
            else:
                length = lenset.pop()
                if self.length_type == LENGTH_NOINIT:
                    self.lengths = length
                    self.length_type = LENGTH_FIXED
                if self.lengths == length:
                    self.length_count += len(lengths)
                else:
                    self._fullLengths()
                    self.processLengths(lengths)
        else:
            if self.length_type == LENGTH_VAR:
                self.lengths.extend(lengths)
            else:
                self._fullLengths()
                self.processLengths(lengths)

    def _fullLengths(self):
        if self.length_type == LENGTH_FIXED:
            self.lengths = [self.lengths] * self.length_count
        elif self.length_type == LENGTH_NOINIT:
            self.lengths = []
        elif self.length_type == LENGTH_REPEAT:
            self.lengths = self.lengths * self.length_count
            self.nparents = 0
        self.length_type = LENGTH_VAR

    def _setDim(self, dim):
        self.dirty = False
        self.dim = dim

    def checkRepeats(self, repeat_length, nparents):
        if self.length_type == LENGTH_FIXED or self.length_type == LENGTH_NOINIT:
            return False

        if self.length_type == LENGTH_REPEAT:
            return self.lengths == repeat_length

        lengths = self.lengths
        if len(lengths) % repeat_length:
            return False

        repeat = lengths[:repeat_length]
        lengths = lengths[repeat_length:]
        while lengths:
            if repeat != lengths[:repeat_length]:
                return False
            lengths = lengths[repeat_length:]
        self.length_count = len(lengths) / repeat_length
        self.lengths = repeat
        self.nparents = nparents
        self.length_type = LENGTH_REPEAT
        return True

    def getDim(self):
        assert self.dirty == False,'DimRep is dirty, has dim equalizer been processed?'
        return self.dim

    def combineWith(self, other):
        self.dirty = True
        if self.length_type == LENGTH_FIXED and other.length_type == LENGTH_FIXED:
            assert self.length_count == other.length_count,'Cannot combine dimreps with unequal lengths'
            self.lengths = self.lengths * other.lengths
        else:
            self._fullLengths()
            other._fullLengths()
            assert len(self.lengths) == len(other.lengths),'Cannot combine dimreps with unequal lenghts'
            self.lengths = list(numpy.multiply(self.lengths, other.lengths))

    def getParentDimReps(self):
        return self.parent_scanner.getDimReps(self.index)


class FixedDimRep(DimRep):
    def __init__(self, parent_scanner, index, dim):
        self.parent_scanner = parent_scanner
        self.index = index

        if(dim.isVariable()):
            self.length_type = LENGTH_VAR
            self.lengths = [] 
        else:
            self.length_type = LENGTH_FIXED
            self.lengths = dim.shape
            #FIXME: give lengths of variable dimensions from data, will also enable repeat var dimensions
        
        self.nparents = 0
        self.has_missing = dim.has_missing
        self.dim = dim
        self.dirty = False


class DimEqualizer(object):

    def __init__(self):
        self.dimreps = []

    def registerDimRep(self, dimrep):
        self.dimreps.append(dimrep)

    def unregisterDimRep(self, dimrep):
        del self.dimreps[self.dimreps.index(dimrep)]

    def _processParents(self, dimrep):
        cur_parents = dimrep.getParentDimReps()
        for parent in cur_parents:
            if parent.dirty:
                self._attachDim(parent)
        
        pos = len(cur_parents)
        repeat_length = 1
        while pos:
            pos -= 1
            parent = cur_parents[pos]
            assert (parent.length_type != LENGTH_NOINIT), 'Unitialized parent should not be possible'
            
            if parent.length_type == LENGTH_VAR: #A VAR as parent means that no repeat could be found, i.e. lengths vary across all parents
                break
            elif parent.length_type == LENGTH_REPEAT:
                repeat_length *= sum(parent.lengths)
                pos -= parent.nparents
            elif parent.lengths == UNDEFINED: # length is UNDEFINED, no further checks possible
                break
            else:
                repeat_length *= parent.lengths

            if dimrep.checkRepeats(repeat_length, len(cur_parents) - pos):
                cur_parents = cur_parents[-dimrep.nparents:]
                break
        return cur_parents

    def _attachDim(self, dimrep):
        assert (dimrep.length_type != LENGTH_NOINIT), 'Unitialized parent should not be possible'

        if dimrep.length_type != LENGTH_FIXED:
            cur_parents = self._processParents(dimrep)
            parents = cur_parents
        else:
            parents = []

        for dr in self.dimreps:
            if dr.dirty is True:
                continue
            if dr.has_missing != dimrep.has_missing:
                continue
            if dimrep.length_type != dr.length_type:
                continue
            if dr.lengths != dimrep.lengths:
                continue
            if dimrep.length_type == LENGTH_FIXED:
                dimrep._setDim(dr.dim)
                return dr.dim
            match_parents = dr.getParentDimReps()[-len(parents):]
            if len(match_parents) != len(cur_parents):
                continue
            for mp, cp in zip(match_parents, cur_parents):
                if mp.dirty:
                    self._attachDim(mp)
                if cp.dim is not mp.dim:
                    break
            else:
                dimrep._setDim(dr.dim)
                return dr.dim
        if dimrep.length_type == LENGTH_FIXED:
            ndim = dimensions.Dim(dimrep.lengths, tuple(), dimrep.has_missing,name="d"+str(self.dimid()))
        else:
            ndim = dimensions.Dim(UNDEFINED, (True,) * len(parents), dimrep.has_missing,name="d"+str(self.dimid()))
        dimrep._setDim(ndim)
        return ndim

    def processDims(self):
        self.dimid = util.seqgen().next
        for dr in self.dimreps:
            if dr.dirty:
                self._attachDim(dr)



class TypeScanner(object):
    need_convert=False
    typecls = rtypes.TypeAny
    good_cls = set()

    def __init__(self, detector):
        self.detector = detector
        self.dimreps = []

    def unregister(self, create_parent=False):
        for d in self.dimreps:
            self.detector.dim_eq.unregisterDimRep(d)

        if create_parent:
            return self.parentcls(self.detector)

    def scan(self, seq):
        return self.detector.objectclss.issubset(self.good_cls)

    def getType(self):
        return self.typecls(self.detector.hasMissing())

    def getAncestorScanners(self):
        if not hasattr(self.__class__, 'ancest_cls_cache'):
            acc = []
            cur_cls = self.__class__.parentcls
            while cur_cls is not None.__class__:
                acc.append(cur_cls)
                cur_cls = cur_cls.parentcls
            self.__class__.ancest_cls_cache = acc
        return self.__class__.ancest_cls_cache
    ancestor_scanners = property(fget=getAncestorScanners)

    def getSubDetector(self, id=0):
        id = 'subdetector_' + str(id)
        if id not in self.__dict__:
            d = Detector(self, self.detector.dim_eq, self.detector.allow_need_convert)
            setattr(self, id, d)
            if self.detector.count_elem:
                d.processSeq([Missing] * self.detector.count_elem)
        return getattr(self, id)

    def getDimReps(self, last_index=None):
        ps = self.detector.parent_scanner
        if last_index is None:
            sdimreps = self.dimreps[:]
        else:
            sdimreps = self.dimreps[:last_index]
        if ps:
            if not self.dimreps:
                return ps.getDimReps()
            res = ps.getDimReps()
            res.extend(sdimreps)
            return res
        return sdimreps

    def getDimRep(self, i):
        if len(self.dimreps) <= i:
            assert len(self.dimreps) == i,'DimReps not requested in order'
            d = DimRep(self, i)
            if self.detector.count_elem:
                d.processLengths(sparse_arrays.FullSparse([Missing] * self.detector.count_elem), has_missing=True)
            self.detector.dim_eq.registerDimRep(d)
            self.dimreps.append(d)
        return self.dimreps[i]

    def reduceDimReps(self, i):
        while len(self.dimreps) > i:
            d = self.dimreps.pop()
            self.dimreps[-1].combineWith(d)
            self.detector.dim_eq.unregisterDimRep(d)
        self.min_dim = i



class AnyScanner(TypeScanner):
    parentcls = None.__class__

    def unregister(self, create_parent=False):
        if create_parent:
            raise RuntimeError, 'Attempting to find parent for AnyScanner'
        return TypeScanner.__init__(self,create_parent)

    def scan(self, seq):
        return True
registerTypeScanner(AnyScanner)
TypeScanner.parentcls = AnyScanner

class TupleScanner(TypeScanner):
    good_cls = set([tuple, MissingType])

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)
        self.max_len = 0
        self.min_len = 999999999
        self.fieldnames = []

    def getType(self):
        if(not len(self.fieldnames) == self.max_len):
            self.fieldnames = ['f' + str(i) for i in xrange(self.max_len)]
        subtypes = tuple([self.getSubDetector(i).getType() for i in xrange(self.max_len)])
        return rtypes.TypeTuple(self.detector.hasMissing(), subtypes, self.fieldnames)
    
    def scan(self, seq):
        if not self.detector.objectclss.issubset(self.good_cls):
            return False

        l = seq.map(len, out_empty=0, has_missing=self.detector.hasMissing(), otype=int)

        maxlen = l.max(out_empty=0)
        minlen = l.min(out_empty=self.min_len)
        self.max_len = max(maxlen, self.max_len)
        self.min_len = min(minlen, self.min_len)

        if(len(seq) == 1 and self.max_len == self.min_len):           
            self.fieldnames = util.find_names(seq[0])
            
        for i in xrange(self.max_len):
            d = self.getSubDetector(i)
            f = operator.itemgetter(i)
            if i < self.min_len:
                subseq = seq.map(f, has_missing=self.detector.hasMissing())
            else:
                subseq = seq.filter_tomissing(l > i).map(f, out_empty=Missing, otype=object, has_missing=True)
            d.processSeq(subseq)
        return True
registerTypeScanner(TupleScanner)


class NamedTupleScanner(TypeScanner):
    bad_cls = set([tuple])

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)
        self.tuple_cls = None

    def getType(self):
        fieldnames = [util.valid_name(name) for name in self.tuple_cls._fields]
        subtypes = tuple([self.getSubDetector(pos).getType() for pos in range(len(fieldnames))])
        return rtypes.TypeTuple(self.detector.hasMissing(), subtypes, fieldnames)

    def scan(self, seq):
        if self.bad_cls.issubset(self.detector.objectclss):
            return False
        for cls in self.detector.objectclss:
            if cls is MissingType:
                continue
            if not self.tuple_cls:
                if not hasattr(cls, '_fields'):
                    return False
                self.tuple_cls = cls
            elif not self.tuple_cls is cls:
                return False

        fieldlen = len(self.tuple_cls._fields)
        for i in xrange(fieldlen):
            d = self.getSubDetector(i)
            f = operator.itemgetter(i)
            subseq = seq.map(f, has_missing=self.detector.hasMissing())
            d.processSeq(subseq)
        return True
registerTypeScanner(NamedTupleScanner)

class RecordDictScanner(TypeScanner):
    good_cls = set([dict, module_types.soap_struct, MissingType])

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)
        self.names = set()

    def getType(self):
        fieldnames = [name for name in self.names]
        subtypes = tuple([self.getSubDetector(name).getType() for name in self.names])
        return rtypes.TypeRecordDict(self.detector.hasMissing(), subtypes, fieldnames)

    def scan(self, seq):
        if not self.detector.objectclss.issubset(self.good_cls):
            return False
        
        names = self.names.copy()
        if dict in self.detector.objectclss:
            assert not module_types.soap_struct in self.detector.objectclss, "dict cannot be mixed with SOAPpy struct type"
            for elem in seq:
                if not (elem is Missing):
                    names.update(elem.keys())
        else:
            assert not dict in self.detector.objectclss, "dict cannot be mixed with SOAPpy struct type"
            for elem in seq:
                if not (elem is Missing):
                    names.update(elem._keys())

        newnames = names - self.names
        for name in newnames:
            if not isinstance(name, basestring) or util.valid_name(name) != name:
                return False
        self.names = names

        if(len(self.names) > 100):
            return False

        for name in self.names:
            d = self.getSubDetector(name)
            f = operator.itemgetter(name)
            def getname(elem):
                try:
                    return elem[name]
                except KeyError:
                    return Missing
            subseq = seq.map(getname, otype=object)
            d.processSeq(subseq)
        return True
registerTypeScanner(RecordDictScanner)


class ContainerScanner(TypeScanner):
    good_cls = set([set, frozenset, MissingType, list, array.array, numpy.ndarray])
    bad_cls = set([tuple, str, unicode, numpy.unicode_, numpy.string_, module_types.soap_struct])

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)
        self.min_dim = None

    def getType(self):
        subtype = self.getSubDetector().getType()
        dims = dimpaths.DimPath(*[self.getDimRep(i).dim for i in xrange(self.min_dim)])
        if self.detector.hasMissing():
            dims[0].has_missing = True

        return dimpaths.dimsToArrays(dims, subtype)

    def scan(self, seq):
        has_missing = self.detector.hasMissing()
        if not self.detector.objectclss.issubset(self.good_cls):
            if not self.detector.objectclss.isdisjoint(self.bad_cls):
                return False

            for cls in self.detector.objectclss:
                if issubclass(cls, tuple):
                    return False

            l = seq.map(operator.isSequenceType, otype=bool, out_empty=True, has_missing=has_missing)
            if not l.all(has_missing=False):
                return False
        
        if(self.min_dim == 1):
            dr = self.getDimRep(0)
            nelems = seq.map(getnelem, otype=object, out_empty=Missing, has_missing=has_missing)
            dr.processLengths(nelems, has_missing=has_missing)
        else:
            shapes = seq.map(getshape, otype=object, out_empty=Missing, has_missing=has_missing)
            shapelens = shapes.map(len, otype=object, out_empty=Missing, has_missing=has_missing)
            min_dim = shapelens.min(has_missing=has_missing)
            max_dim = shapelens.max(has_missing=has_missing)

            if(min_dim is Missing or max_dim is Missing):
                if(self.min_dim):
                    for i in xrange(self.min_dim):
                        dr = self.getDimRep(i)
                        dr.processLengths(shapelens, has_missing=has_missing)
            else:
                if self.min_dim is None:
                    self.min_dim = min_dim
                else:
                    if min_dim < self.min_dim:
                        self.reduceDimReps(min_dim)

                for i in xrange(self.min_dim):
                    dr = self.getDimRep(i)
                    if i == self.min_dim and self.min_dim < max_dim:
                        red = numpy.multiply.reduce
                        def reduceshape(shape):
                            return red(shape[i:])
                        f = reduceshape
                    else:
                        f = operator.itemgetter(i)
                    nelems = shapes.map(f, otype=object, out_empty=Missing, has_missing=has_missing)
                    dr.processLengths(nelems.ravel(), has_missing=has_missing)
        
        d = self.getSubDetector()
        if(self.min_dim == 1 and not has_missing):
            d.processSeq(list(chain(*seq.ravel())))
        else: 
            for subseq in seq.ravel():
                if not (subseq is Missing):
                    d.processSeq(subseq)
        return True

def getshape(elem):
    try:
        return elem.shape
    except: 
        return (len(elem),)

def getnelem(elem):
    try:
        return len(elem.ravel())
    except:
        return len(elem)
registerTypeScanner(ContainerScanner)


class OuterContainerScanner(ContainerScanner):
    parentcls = None.__class__

    def __init__(self, detector, dims):
        ContainerScanner.__init__(self,detector)
        self.dimreps = [FixedDimRep(self,pos,dim) for pos,dim in enumerate(dims)]

    def scan(self,seq):
        d = self.getSubDetector()
        d.processSeq(seq)
        return True

    def getType(self):
        return self.getSubDetector().getType()


class SetScanner(ContainerScanner):
    parentcls = ContainerScanner
    good_cls = set([set, frozenset, MissingType])

    def getType(self):
        subtype = self.getSubDetector().getType()
        dim = dimensions.Dim(UNDEFINED, (True,) * len(self.getDimReps(0)), self.detector.hasMissing())
        dims = dimpaths.DimPath(dim)
        return rtypes.TypeSet(self.detector.hasMissing(), dims, (subtype,))

    def unregister(self, create_parent=False):
        parent = ContainerScanner.unregister(self, create_parent)
        if create_parent:
            parent.min_dim = 1
            parent.dimreps = self.dimreps
        return parent

    def scan(self, seq):
        has_missing = self.detector.hasMissing()
        if not self.detector.objectclss.issubset(self.good_cls):
            return False
        
        d = self.getSubDetector()
        for subseq in seq.ravel():
            if not (subseq is Missing):
                d.processSeq(subseq)
        return True
registerTypeScanner(SetScanner)


class StringScanner(TypeScanner):
    good_cls = set([str, unicode, MissingType, numpy.string_, numpy.unicode_])
    unicode_cls = set([unicode, numpy.unicode_])

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)
        self.max_nchars = 0

    def getType(self):
        if self.unicode_cls.isdisjoint(self.detector.objectclss):
            ntype = rtypes.TypeBytes
        else:
            ntype = rtypes.TypeString
        if self.max_nchars < 32:
            d = dimensions.Dim(self.max_nchars, tuple(), self.detector.hasMissing())
        else:
            d = dimensions.Dim(UNDEFINED, (True,) * len(self.getDimReps(0)), self.detector.hasMissing())

        dims = dimpaths.DimPath(d)
        return ntype(self.detector.hasMissing(), dims)

    def scan(self, seq):
        has_missing = self.detector.hasMissing()
        if not self.detector.objectclss.issubset(self.good_cls):
            return False
        if(self.max_nchars < 32):
            max_nchars = seq.map(len, otype=int, out_empty=-1, has_missing=has_missing).max(has_missing=False)
            self.max_nchars = max(self.max_nchars, max_nchars)
        return True
registerTypeScanner(StringScanner)

class StringRealScanner(StringScanner):
    need_convert=True
    parentcls=StringScanner
    missing_str = set(["", "NA", "N/A","NaN", "nan", "--", "?", "null"])
    def __init__(self, detector):
        StringScanner.__init__(self, detector)
        self.has_missing = False
    
    def unregister(self, create_parent=False):
        res = super(StringRealScanner,self).unregister(create_parent)
        if create_parent:
            res.max_nchars = self.max_nchars
        return res

    def getType(self):
        return rtypes.TypeReal64(self.detector.hasMissing() or self.has_missing)

    def scan(self, seq):
        res = StringScanner.scan(self, seq)
        if res:
            for elem in seq.ravel():
                try:
                    float(elem)
                except (ValueError, TypeError):
                    if elem.__class__ in missing_cls or elem in self.missing_str:
                        self.has_missing = True
                    else:
                        return False
        return res
registerTypeScanner(StringRealScanner)

class StringIntScanner(StringRealScanner):
    parentcls=StringRealScanner
    
    def unregister(self, create_parent=False):
        res = super(StringIntScanner,self).unregister(create_parent)
        if create_parent:
            res.has_missing = self.has_missing
        return res

    def getType(self):
        return rtypes.TypeInt64(self.detector.hasMissing() or self.has_missing)

    def scan(self, seq):
        res = StringScanner.scan(self, seq)
        if res:
            for elem in seq.ravel():
                try:
                    int(elem)
                except (ValueError, TypeError):
                    if elem.__class__ in missing_cls or elem in self.missing_str:
                        self.has_missing = True
                    else:
                        return False
        return res
registerTypeScanner(StringIntScanner)

class SliceScanner(TypeScanner):
    __doc__ = 'Slice scanner'
    typecls = rtypes.TypeSlice
    good_cls = set((slice, MissingType))

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)

    def getType(self):
        if not self.detector.hasMissing():
            pass
        has_missing = self.detector.hasMissing()
        return self.typecls(has_missing)

    def scan(self, seq):
        if not self.detector.objectclss.issubset(self.good_cls):
            return False
        return True
registerTypeScanner(SliceScanner)

class NumberScanner(TypeScanner):
    __doc__ = 'Number scanner, accepts all number objects.'
    typecls = rtypes.TypeNumber
    good_cls = set((bool, float, complex, long, int, numpy.int8, numpy.int16, numpy.int32, numpy.int64, numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64, numpy.float32, numpy.float64, numpy.complex64, numpy.complex128, numpy.bool_, MissingType))

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)
        self.has_nan = False

    def getType(self):
        if not self.detector.hasMissing():
            pass
        has_missing = self.detector.hasMissing() or self.has_nan
        return self.typecls(has_missing)

    def scan(self, seq):
        if not self.detector.objectclss.issubset(self.good_cls):
            return False
        if not self.has_nan:
            self.has_nan = numpy.nan in seq
        return True
registerTypeScanner(NumberScanner)


class RealScanner(NumberScanner):
    __doc__ = 'Real scanner, accepts all real and integer objects.'
    parentcls = NumberScanner
    typecls = rtypes.TypeReal64
    good_cls = set((float, bool, numpy.float32, numpy.float64, MissingType, long, int, numpy.int8, numpy.int16, numpy.int32, numpy.int64, numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64, numpy.bool_))
registerTypeScanner(RealScanner)


class IntegerScanner(TypeScanner):
    parentcls = RealScanner
    ustepvals = [256, 65536, 4294967296L, 18446744073709551616L]
    uinttypes = [rtypes.TypeUInt8, rtypes.TypeUInt16, rtypes.TypeUInt32, rtypes.TypeUInt64]
    istepvals = [128, 32768, 2147483648L, 9223372036854775808L]
    inttypes = [rtypes.TypeInt8, rtypes.TypeInt16, rtypes.TypeInt32, rtypes.TypeInt64]
    
    good_cls = set((bool, long, int, numpy.int8, numpy.int16, numpy.int32, numpy.int64, numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64, numpy.bool_, MissingType))
    
    numpy_minmax = {numpy.dtype('bool'): (0, 1), numpy.dtype('int8'): (-128, 127), numpy.dtype('uint8'): (0, 255), numpy.dtype('int16'): (-32768, 32767), numpy.dtype('uint16'): (0, 65535), numpy.dtype('int32'): (-2147483648, 2147483647), numpy.dtype('uint32'): (0, 4294967295L), numpy.dtype('int64'): (-9223372036854775808L, 9223372036854775807L), numpy.dtype('uint64'): (0, 18446744073709551615L)}
    array_minmax = {'b': (-128, 127), 'B': (0, 255), 'h': (-32768, 32767), 'H': (0, 65535), 'i': (-2147483648, 2147483647), 'I': (0, 4294967295L), 'l': (-2147483648, 2147483647), 'L': (0, 4294967295L)}

    def __init__(self, detector):
        TypeScanner.__init__(self, detector)
        self.max_val = 0
        self.min_val = 0

    def getType(self):
        out_type = rtypes.TypeInteger
        if self.min_val >= 0:
            for stepval, rtype in zip(self.ustepvals, self.uinttypes):
                if self.max_val < stepval:
                    out_type = rtype
                    break
        else:
            for stepval, rtype in zip(self.istepvals, self.inttypes):
                if -self.min_val <= stepval and self.max_val < stepval:
                    out_type = rtype
                    break
        return out_type(self.detector.hasMissing())

    def scan(self, seq):
        if not self.detector.objectclss.issubset(self.good_cls):
            return False
        
        if isinstance(seq, numpy.ndarray) and not seq.dtype == object:
            minmax = self.numpy_minmax[seq.dtype]
        elif isinstance(seq, array.array):
            minmax = self.array_minmax[seq.typecode]
        elif len(seq) == 0:
            minmax = (0, 0)
        else:
            minmax = (0,0)
            for cls in self.detector.objectclss:
                if cls is MissingType:
                    continue
                tminmax = self.numpy_minmax[numpy.dtype(int)]
                minmax = (min(tminmax[0],minmax[0]),max(tminmax[1],minmax[1]))

        self.min_val = min(minmax[0], self.min_val)
        self.max_val = max(minmax[1], self.max_val)
        return True
registerTypeScanner(IntegerScanner)

class BoolScanner(TypeScanner):
    parentcls = IntegerScanner
    typecls = rtypes.TypeBool
    good_cls = set((numpy.bool_, bool, MissingType))
registerTypeScanner(BoolScanner)



import itertools
import manager
import create_graph
import python_peephole
import serialize_exec

from ..utils import util
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, F_CACHE
from ..constants import *
_delay_import_(globals(),"..query_graph")
_delay_import_(globals(),"..ops")
_delay_import_(globals(),"..itypes","dimpaths","rtypes")
_delay_import_(globals(),"..repops_funcs")

class Transposer(VisitorFactory(prefixes=("visit",), flags=NF_ERROR), manager.Pass):
    after=set([python_peephole.PythonPeepHole])
    before=set([serialize_exec.SerializeExec])

    @classmethod
    def run(cls, query, run_manager):
        self = cls()
        self.graph = run_manager.pass_results[create_graph.CreateGraph]

        for node in list(self.graph.nodes):
            self.visit(node)
    
    def visitNode(self, node):
        pass

    def visitFilterOp(self, node):

        #check 1: all dims fixed, more than one dim
        if not len(node.dims) > 1:
            return
        if not all([not d.isVariable() for d in node.dims]):
            return

        #check 2: type dim dependent
        if not node.type.__class__ is rtypes.TypeArray:
            return

        d = node.type.dims[0]
        if not d.isVariable():
            return

        #check 3: type dim not dependent on all dims
        if not (len(d.dependent) < len(node.dims) or not all(d.dependent)):
            return

        #check 5: type dim aggregated, single use, aggretation type does not include dim
        for n in self.graph.walkUseContig(node, include_first=False):
            if isinstance(n, (ops.UnpackArrayOp, ops.PackArrayOp)):
                pass
            elif isinstance(n, ops.UnaryFuncAggregateOp):
                last = self.graph.getDataEdge(n)
                packdim = last.source.dims[-n.packdepth]
                if not packdim == d:
                    return #FIXME: possible to have multiple unaryfuncaggregates...
                if n.type.__class__ == rtypes.TypeArray:
                    if n.type.dims[0] in node.dims:
                        return 
                aggnode = n
                break
            else:
                return
        else:#ran out of contig
            return

        #determine permute dims
        permdim = set()
        for dep, dim in itertools.izip_longest(d.dependent, node.dims[::-1], fillvalue=False):
            if not dep:
                permdim.add(dim)

        
        #insert pre permute
        constraint = self.graph.getDataEdge(node, name="constraint").source
        data = self.graph.getDataEdge(node).source
        


        #check 6: all constraint insert dims correspond to perm dims
        constraint_insert_dims = []
        tdims = list(node.dims)
        for n in self.graph.walkSourceContig(constraint):
                if isinstance(n, ops.InsertDimOp):
                    if not tdims[n.matchpoint] in permdim:
                        return
                    constraint_insert_dims.append(tdims[n.matchpoint])
                    del tdims[n.matchpoint]
                else:
                    constraint_source_node = n
                    assert tuple(tdims) == tuple(n.dims), "Dims do not match after deinstertion"
                    break
        else:
                return #ran out of contig
       
        #limit permdims to constraint insert dims
        permdim = permdim & set(constraint_insert_dims)

        #check 7: all data insert dims correspond to non perm dims
        data_insert_dims = []
        data_insert_dims_bc = []
        tdims = list(node.dims)
        for n in self.graph.walkSourceContig(data):
                if isinstance(n, ops.InsertDimOp):
                    if tdims[n.matchpoint] in permdim:
                        return
                    data_insert_dims.append(tdims[n.matchpoint])
                    data_insert_dims_bc.append(n.dims[n.matchpoint])
                    del tdims[n.matchpoint]
                else:
                    data_source_node = n
                    assert tuple(tdims) == tuple(n.dims), "Dims do not match after deinsertion"
                    break
        else:
                return #ran out of contig

        #add unpack node
        while(data_source_node.type.__class__ == rtypes.TypeArray and
                data_source_node.dims[0] in permdim):
                newnode = ops.UnpackArrayOp(data_source_node)
                self.graph.addNode(newnode)
                self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
                data_source_node = newnode

        #determine permute_idxs in 
        before = []
        after = []
        for pos, dim in enumerate(data_source_node.dims):
            if dim in permdim:
                after.append(pos)
            else:
                before.append(pos)
        permute_idxs = before + after
        
        #add permute node
        newnode = ops.PermuteDimsOp(data_source_node, permute_idxs)
        self.graph.addNode(newnode)
        self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
        data_source_node = newnode
                
        #add pack node of perm dims
        newnode = ops.PackArrayOp(data_source_node, len(after)+1)
        self.graph.addNode(newnode)
        self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
        data_source_node = newnode
        
        #redo inserts
        insertpos = 0
        for dim in constraint_source_node.dims:
                if not dim in data_source_node.dims:
                    bcdim = data_insert_dims_bc[data_insert_dims.index(dim)]
                    newnode = ops.InsertDimOp(data_source_node, insertpos, bcdim)
                    self.graph.addNode(newnode)
                    self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
                    data_source_node = newnode
                insertpos = insertpos + 1
                    
        #remove constraint inserts
        cnode = constraint
        while isinstance(cnode, ops.InsertDimOp):
                cnode = self.graph.remove_unaryop(cnode)
        assert cnode == constraint_source_node, "Node mismatch" 
        

        #reconstruct filter op
        newnode = ops.FilterOp(data_source_node, constraint_source_node, d)
        self.graph.addNode(newnode)
        self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
        self.graph.addEdge(query_graph.ParamEdge(constraint_source_node, newnode,"constraint"))
        data_source_node = newnode

        #replace bc dims
        ndims = []
        for dim in data_source_node.dims:
            if dim in data_insert_dims_bc:
                bcdim = data_insert_dims[data_insert_dims_bc.index(dim)]
                ndims.append(bcdim)
            else:
                ndims.append(dim)
        data_source_node.dims = dimpaths.DimPath(*ndims)

        #do unpack
        newnode = ops.UnpackArrayOp(data_source_node, len(after) + 1)
        self.graph.addNode(newnode)
        self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
        data_source_node = newnode
        

        #redo aggregate
        outparam = repops_funcs.Param(aggnode.name, aggnode.type)
        newnode = ops.UnaryFuncAggregateOp(aggnode.funcname, aggnode.sig, outparam, 
                                            len(after) + 1, data_source_node, **aggnode.kwargs)
        self.graph.addNode(newnode)
        self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
        data_source_node = newnode
                                            

        #determine permute_idxs out
        permute_idxs = [data_source_node.dims.index(dim) for dim in aggnode.dims]


        #insert post permute
        newnode = ops.PermuteDimsOp(data_source_node, permute_idxs)
        self.graph.addNode(newnode)
        self.graph.addEdge(query_graph.ParamEdge(data_source_node, newnode,"slice"))
        data_source_node = newnode

        #remove target nodes aggnode
        edges = self.graph.edge_source[aggnode]
        for edge in list(edges):
            self.graph.dropEdge(edge)
            edge.source = data_source_node
            self.graph.addEdge(edge)

        self.graph.pruneGraph()
            




_delay_import_(globals(),"repops")
from utils import util

class QueryContext(object):
    """Context object of a query. Encloses representor
    with fixate operation."""
    def __init__(self, representor, args={}, endpoint=True):
        if(endpoint):
            self.root = repops.Fixate(representor)
        else:
            self.root = repops.Gather(representor)
            
        self.args = args




import copy
import weakref

from ..utils import util
from ..constants import *
import itertools

_delay_import_(globals(),"dimpaths")


#pylint: disable-msg=E1101
dimid = util.seqgen().next
def getNewDimid():
    return dimid()
class Dim(object):
    """Class representing a dimension."""
    __slots__ = ['id', 'name', 'shape', 'dependent', 'has_missing','redim_cache','__weakref__']
   
    def __reduce__(self):
        return (Dim, (self.shape, self.dependent, self.has_missing, self.id, self.name))

    def __init__(self, shape, dependent=tuple(), has_missing=False, did=None, name=None):
        """Creates dimension class.

        :param shape: shape of dimension. Can be integer >= 0 or UNDEFINED
        :param dependent: tuple of bools, indicating on which parent dimension this dim is dependent
        :param has_missing: indicates if there are Missing values
        :param did:   id of dimension, optional
        """
        if(not did):
            self.id = getNewDimid()
        else:
            self.id = did

        assert isinstance(self.id, int), "ID should be a integer"

        if(name is None):
            self.name = "d" + str(self.id)
        else:
            assert name.lower() == name, "Dimension name should be lower case"
            self.name = name
        
        self.has_missing = has_missing
        
        assert (isinstance(dependent, tuple) and all([isinstance(elem,bool) for elem in dependent])), "Variable should be a tuple of bools"
        while(dependent and dependent[-1] is False): #remove False at end, as they do not contribute
            dependent = dependent[:-1]

        self.dependent = dependent
        self.shape = shape

    def isVariable(self):
        return len(self.dependent) > 0

    def __len__(self):
        return len(self.dependent)
    
    def __iter__(self):
        return self.dependent.__iter__()

    def _getRedimCache(self):
        try:
            return self.redim_cache
        except AttributeError:
            self.redim_cache = weakref.WeakValueDictionary()
            return self.redim_cache
            

    def removeDepDim(self, pos, elem_specifier):
        assert pos >= 0, "Position should not be smaller than 0"
        if(pos >= len(self.dependent)):
            return self

        key = (0, pos, elem_specifier)
        redim_cache = self._getRedimCache()
        if(not key in redim_cache):
            nself = self.copy(reid=self.dependent[pos])
            ndependent = self.dependent[:pos] + self.dependent[(pos + 1):]
            while(ndependent and ndependent[-1] is False):
                ndependent = ndependent[:-1]
            nself.dependent = ndependent
            redim_cache[key] = nself

        return redim_cache[key]
    
    def updateDepDim(self, pos, ndim):
        assert pos >= 0, "Position should not be smaller than 0"
        if(pos >= len(self.dependent)):
            return self

        if(not isinstance(ndim,tuple)):
            ndim = (ndim,)
        assert ndim, "No new dimensions specified"
        assert len(ndim) == 1, "Trap: updatedepdim with multiple dims"

        key = (1,pos, tuple([nd.id for nd in ndim]))
        redim_cache = self._getRedimCache()
        if(not key in redim_cache):
            nself = self.copy(reid=self.dependent[pos])
            if nself.dependent[pos]:
                ndependent = tuple([a or b for a,b in itertools.izip_longest(ndim[0].dependent, nself.dependent[(pos+1):],fillvalue=False)])
                nself.dependent = nself.dependent[:(pos + 1)] + ndependent
            redim_cache[key] = nself
        return redim_cache[key]
    
    def insertDepDim(self, pos, ndim):
        assert pos >= 0, "Position should not be smaller than 0"
        if(pos >= len(self.dependent)):
            return self

        redim_cache = self._getRedimCache()
        key = (2, pos, ndim.id)
        if(not key in redim_cache):
            nself = self.copy()
            nself.dependent = self.dependent[:pos] + (False,) + self.dependent[pos:]
            redim_cache[key] = nself
        return redim_cache[key]

    def changeDependent(self, dep, ndims):
        assert len(dep) == len(ndims), "Number of flags should be equal to number of dims when setting dependent dims"
        dep = tuple(dep)
        while(dep and dep[-1] is False):
            dep = dep[:-1]

        if(dep == self.dependent):
            return self

        redim_cache = self._getRedimCache()
        key = (3, tuple([ndim.id for d, ndim in zip(dep,ndims) if d]))

        if(not key in redim_cache):
            nself = self.copy(reid=True)
            nself.dependent = dep
            redim_cache[key] = nself
        nself = redim_cache[key]

        if(not nself.dependent == dep):
            nself = self.copy(reid=False)
            nself.dependent = dep

        return nself
       

    def copy(self, reid=False):
        """Returns a copy of this object"""
        res = copy.copy(self)
        if(reid):
            res.id = getNewDimid()
        return res

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and  self.id == other.id)

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        if(self.dependent):
            if(self.shape == 1):
                res = self.name + ":?"
            else:
                res = self.name + ":~"
        elif(self.shape == UNDEFINED):
            res = self.name + ":*"
        else:
            res = self.name + ":" + str(self.shape)
        if(self.has_missing):
            res += "?"
        return res


    def merge(self, other):
        if self == other:
            return self
        if self.shape == other.shape and self.shape != UNDEFINED and self.name == other.name and not self.dependent and not other.dependent and not self.has_missing and not other.has_missing:
            return self
        redim_cache = self._getRedimCache()
        key = (self,other)
        if not key in redim_cache:
            if(self.shape == UNDEFINED or other.shape == UNDEFINED):
                rshape = UNDEFINED
            else:
                rshape = max(self.shape, other.shape)

            ndep = tuple([ldep or rdep for ldep, rdep in itertools.izip_longest(self.dependent, other.dependent,fillvalue=False)])

            if(self.name == other.name):
                nname = self.name
            else:
                nname = self.name + "_" + other.name
            
            ndim = Dim(rshape, ndep, self.has_missing or other.has_missing, name=nname)
            redim_cache[key] = ndim
        return redim_cache[key]

            

def toCommonDim(name, dims):
    shapes = set()
    dependent = set()
    for dim in dims:
        shapes.add(dim.shape)
        dependent.add(dim.dependent)
    shapes.discard(UNDEFINED)
    assert len(shapes) <= 1, "Dims with different shape cannot be cast to each other"
    if not shapes:
        shape = UNDEFINED
    else:
        shape = shapes.pop()

    assert len(dependent) == 1, "Dims with different dependencies cannot be cast to each other"
    dependent = dependent.pop()
    return Dim(shape, dependent, name=name)


        
        



def getTerminalSize():
    """
    returns (lines:int, cols:int)
    """
    import os, struct
    def ioctl_GWINSZ(fd):
        import fcntl, termios
        return struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))

    # try stdin, stdout, stderr
    for fd in (0, 1, 2):
        try:
            return ioctl_GWINSZ(fd)
        except:
            pass
    
    # try os.ctermid()
    try:
        fd = os.open(os.ctermid(), os.O_RDONLY)
        try:
            return ioctl_GWINSZ(fd)
        finally:
            os.close(fd)
    except:
       pass

    # try `stty size`
    try:
        return tuple(int(x) for x in os.popen("stty size", "r").read().split())
    except:
        pass
    
    # try environment variables
    try:
        return tuple(int(os.getenv(var)) for var in ("LINES", "COLUMNS"))
    except:
        pass
    
    # i give up. return default.
    return (25, 80)

